package net.sf.csutils.core.model;

import javax.xml.XMLConstants;


/**
 * Kind of a qualified name, but without the restrictions of
 * {@code NCName} on the local part.
 */
public class QName {
	private static final String[] PREFIXES = new String[]{
		XMLConstants.DEFAULT_NS_PREFIX
	};
	private static final String[] URIS = new String[]{
		"" // XMLConstants.NULL_NS_URI
	};
	private String uri, localPart, prefix;

	/**
	 * Creates a new instance with the default namespace
	 * and no prefix.
	 * @param pLocalPart The qualified names local part.
	 */
	public QName(String pLocalPart) {
		this("", pLocalPart, "");
	}

	/**
	 * Creates a new instance without prefix.
     * @param pUri The qualified names namespace URI.
     * @param pLocalPart The qualified names local part.
	 */
	public QName(String pUri, String pLocalPart) {
		this(pUri, pLocalPart, "");
	}

	/**
	 * Creates a new instance with the given namespace
	 * URI, local part, and prefix.
     * @param pUri The qualified names namespace URI.
     * @param pLocalPart The qualified names local part.
     * @param pPrefix The qualified names prefix.
	 */
	public QName(String pUri, String pLocalPart, String pPrefix) {
		if (pUri == null) {
			throw new IllegalArgumentException("The namespace URI must not be null.");
		}
		if (pLocalPart == null) {
			throw new IllegalArgumentException("The local part must not be null.");
		}
		if (pPrefix == null) {
			throw new IllegalArgumentException("The namespace prefix must not be null.");
		}
		uri = pUri;
		localPart = pLocalPart;
		prefix = pPrefix;
	}

	/**
	 * Returns the namespace URI.
	 * @return The qualified names namespace URI.
	 */
	public String getNamespaceURI() {
		return uri;
	}

	/**
	 * Returns the local part.
     * @return The qualified names local part.
	 */
	public String getLocalPart() {
		return localPart;
	}

	/**
	 * Returns the namespace prefix.
     * @return The qualified names prefix.
	 */
	public String getPrefix() {
		return prefix;
	}

	public static QName valueOf(String pValue) {
		if (pValue.length() == 0) {
			throw new IllegalArgumentException("Empty QName is an illegal argument.");
		}
		if (pValue.charAt(0) == '{') {
			int offset = pValue.indexOf('}');
			if (offset == -1) {
				throw new IllegalArgumentException("Invalid QName, missing '}': " + pValue);
			}
			if (offset+1 == pValue.length()) {
				throw new IllegalArgumentException("Invalid QName, missing local part: " + pValue);
			}
			final String uri = pValue.substring(1, offset);
			final String localPart = pValue.substring(offset+1);
			return new QName(uri, localPart);
		}
		int offset = pValue.indexOf(':');
		if (offset == -1) {
			return new QName(pValue);
		}
		if (offset+1 == pValue.length()) {
			throw new IllegalArgumentException("Invalid QName, missing local part: " + pValue);
		}
		final String prefix = pValue.substring(0, offset);
		final String localPart = pValue.substring(offset+1);
		for (int i = 0;  i < PREFIXES.length;  i++) {
			if (PREFIXES[i].equals(prefix)) {
				return new QName(URIS[i], localPart, prefix);
			}
		}
		throw new IllegalArgumentException("Invalid QName, unknown namespace prefix: " + pValue);
	}

	public static String toString(QName pQName) {
		final String uri = pQName.getNamespaceURI();
		for (int i = 0;  i < PREFIXES.length;  i++) {
			if (PREFIXES[i].equals(uri)) {
				return uri;
			}
		}
		throw new IllegalArgumentException("Invalid QName, unknown namespace prefix: " + pQName);
	}

	@Override
	public String toString() {
		final String nsuri = getNamespaceURI();
		return XMLConstants.NULL_NS_URI.equals(nsuri) ?
				getLocalPart() : ('{' + nsuri + '}' + getLocalPart());
	}

	@Override
	public int hashCode() {
		return 31 * (31 + localPart.hashCode()) + uri.hashCode();
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		QName other = (QName) obj;
		return localPart.equals(other.localPart) && uri.equals(other.uri);
	}
}
