package net.sf.csutils.core.model.impl;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.Reader;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.xml.bind.JAXBException;
import javax.xml.registry.JAXRException;
import javax.xml.stream.XMLEventReader;
import javax.xml.stream.XMLStreamReader;
import javax.xml.transform.Source;
import javax.xml.transform.sax.SAXSource;
import javax.xml.transform.stream.StreamSource;

import net.sf.csutils.core.model.QName;
import net.sf.csutils.core.model.ROAttribute;
import net.sf.csutils.core.model.ROClassification;
import net.sf.csutils.core.model.ROFile;
import net.sf.csutils.core.model.ROMetaModel;
import net.sf.csutils.core.model.RORelation;
import net.sf.csutils.core.model.ROSlot;
import net.sf.csutils.core.model.ROType;
import net.sf.csutils.core.model.vo.RORelation.TargetTypes;
import net.sf.csutils.core.model.vo.ROTypeReference;
import net.sf.csutils.core.utils.JaxbReader;

import org.w3c.dom.Node;
import org.xml.sax.InputSource;

/**
 * The registry object model writer is used to deserialize
 * registry meta models from XML.
 */
public class ROMetaModelReader {
	private static interface LateBinding {
		void bind(ROMetaModel pModel) throws JAXRException;
	}

	private static final JaxbReader<net.sf.csutils.core.model.vo.ROMetaModel> jaxbReader = new JaxbReader<net.sf.csutils.core.model.vo.ROMetaModel>(net.sf.csutils.core.model.vo.ROMetaModel.class);
	private List<LateBinding> lateBindings = new ArrayList<LateBinding>();

	protected ROSlot read(net.sf.csutils.core.model.vo.ROSlot pBean) {
		final StaticROSlot attr = new StaticROSlot();
		map(pBean, attr);
		attr.setType(ROAttribute.Type.valueOf(pBean.getType().value()));
		return attr;
	}

	private void map(net.sf.csutils.core.model.vo.ROAttribute pBean,
			final StaticROAttribute pAttr) {
		pAttr.setName(pBean.getName());
		final String maxOccurs = pBean.getMaxOccurs();
		pAttr.setMaxOccurs("unbounded".equals(maxOccurs) ? -1 : Integer.parseInt(maxOccurs));
		pAttr.setMinOccurs(pBean.getMinOccurs());
	}

    protected ROClassification read(final net.sf.csutils.core.model.vo.ROClassification pBean) {
        final StaticROClassification attr = new StaticROClassification();
        map(pBean, attr);
        attr.setTaxonomy(pBean.getTaxonomy());
        attr.setAttributeKey(pBean.getAttributeKey());
        return attr;
    }

    protected ROFile read(final net.sf.csutils.core.model.vo.ROFile pBean) {
        final StaticROFile attr = new StaticROFile();
        map(pBean, attr);
        attr.setAttributeKey(pBean.getAttributeKey());
        return attr;
    }

    protected RORelation read(final net.sf.csutils.core.model.vo.RORelation pBean) {
		final StaticRORelation attr = new StaticRORelation();
		map(pBean, attr);
		String assocType = pBean.getAssociationType();
		if (assocType == null) {
			assocType = pBean.getName();
		}
		attr.setAssociationType(pBean.getAssociationType());
		attr.setAttributeKey(pBean.getAttributeKey());
		lateBindings.add(new LateBinding(){
			@Override
			public void bind(ROMetaModel pModel) throws JAXRException {
				final List<ROType> targetTypeList = new ArrayList<ROType>();
				final TargetTypes targetTypesBean = pBean.getTargetTypes();
				if (targetTypesBean != null) {
					for (ROTypeReference ref : targetTypesBean.getTargetType()) {
						final QName targetTypeQName = QName.valueOf(ref.getQName());
						final ROType type = pModel.getROType(targetTypeQName);
						if (type == null) {
							throw new JAXRException("Unknown target type " + targetTypeQName);
						}
						targetTypeList.add(type);
					}
				}
				attr.setTargetTypes(targetTypeList);
			}
		});
		return attr;
	}

	protected ROType read(net.sf.csutils.core.model.vo.ROType pBean) {
		final StaticROType type = new StaticROType();
		type.setQName(QName.valueOf(pBean.getQName()));
		final String queryName = pBean.getQueryName();
		if (queryName != null) {
		    type.setQueryName(QName.valueOf(queryName));
		}
		final Map<String,ROAttribute> attrs = new HashMap<String,ROAttribute>();
		final net.sf.csutils.core.model.vo.ROType.Attributes attrsBean = pBean.getAttributes();
		if (attrsBean != null) {
			for (net.sf.csutils.core.model.vo.ROAttribute attrBean : attrsBean.getAttributeOrRelationOrClassification()) {
				final ROAttribute attr;
				if (attrBean instanceof net.sf.csutils.core.model.vo.ROClassification) {
					attr = read((net.sf.csutils.core.model.vo.ROClassification) attrBean);
				} else if (attrBean instanceof net.sf.csutils.core.model.vo.RORelation) {
				    attr = read((net.sf.csutils.core.model.vo.RORelation) attrBean);
				} else if (attrBean instanceof net.sf.csutils.core.model.vo.ROFile) {
				    attr = read((net.sf.csutils.core.model.vo.ROFile) attrBean);
				} else {
					attr = read((net.sf.csutils.core.model.vo.ROSlot) attrBean);
				}
				attrs.put(attr.getName(), attr);
			}
		}
		type.setAttributes(attrs);
		return type;
	}

	protected ROMetaModel read(net.sf.csutils.core.model.vo.ROMetaModel pBean) throws JAXRException {
		final StaticROMetaModel model = new StaticROMetaModel();
		final Map<QName,ROType> types = new HashMap<QName,ROType>();
		final net.sf.csutils.core.model.vo.ROMetaModel.RoTypes typesBean = pBean.getRoTypes();
		if (typesBean != null) {
			for (net.sf.csutils.core.model.vo.ROType typeBean : typesBean.getRoType()) {
				final ROType type = read(typeBean);
				types.put(type.getQName(), type);
			}
		}
		model.setROTypes(types);
		for (LateBinding lateBinding : lateBindings) {
			lateBinding.bind(model);
		}
		return model;
	}

	/**
	 * Reads and deserializes a meta model from the XML stream provided
	 * by the given {@link InputStream}.
	 * @param pStream The stream, from which to read the meta model.
	 * @return The created meta model.
	 * @throws JAXRException The operation failed.
	 */
	public ROMetaModel read(InputStream pStream) throws JAXRException {
		return read(new StreamSource(pStream));
	}

	/**
	 * Reads and deserializes a meta model from the XML stream provided
	 * by the given {@link File}.
     * @param pFile The file, from which to read the meta model.
     * @return The created meta model.
     * @throws JAXRException The operation failed.
	 */
	public ROMetaModel read(File pFile) throws JAXRException {
		return read(new StreamSource(pFile));
	}

	/**
	 * Reads and deserializes a meta model from the XML stream provided
	 * by the given {@link InputSource}.
     * @param pSource The SAX source, from which to read the meta model.
     * @return The created meta model.
     * @throws JAXRException The operation failed.
	 */
	public ROMetaModel read(InputSource pSource) throws JAXRException {
		return read(new SAXSource(pSource));
	}

	/**
	 * Reads and deserializes a meta model from the XML stream provided
	 * by the given {@link Reader}.
     * @param pReader The reader, from which to read the meta model.
     * @return The created meta model.
     * @throws JAXRException The operation failed.
	 */
	public ROMetaModel read(Reader pReader) throws JAXRException {
		return read(new StreamSource(pReader));
	}

	/**
	 * Reads and deserializes a meta model from the XML stream provided
	 * by the given {@link Node}.
     * @param pNode The node, from which to read the meta model.
     * @return The created meta model.
     * @throws JAXRException The operation failed.
	 */
	public ROMetaModel read(Node pNode) throws JAXRException {
		try {
			final net.sf.csutils.core.model.vo.ROMetaModel metaModel = jaxbReader.read(pNode);
			return read(metaModel);
		} catch (JAXBException e) {
			throw new JAXRException(e);
		}
	}

	/**
	 * Reads and deserializes a meta model from the XML stream provided
	 * by the given {@link Source}.
     * @param pSource The transformer source, from which to read the meta model.
     * @return The created meta model.
     * @throws JAXRException The operation failed.
	 */
	public ROMetaModel read(Source pSource) throws JAXRException {
		try {
			final net.sf.csutils.core.model.vo.ROMetaModel metaModel = jaxbReader.read(pSource);
			return read(metaModel);
		} catch (JAXBException e) {
			throw new JAXRException(e);
		}
	}

	/**
	 * Reads and deserializes a meta model from the XML stream provided
	 * by the given {@link URL}.
     * @param pURL The URL, from which to read the meta model.
     * @return The created meta model.
     * @throws JAXRException The operation failed.
	 */
	public ROMetaModel read(URL pURL) throws JAXRException {
		try {
			final net.sf.csutils.core.model.vo.ROMetaModel metaModel = jaxbReader.read(pURL);
			return read(metaModel);
		} catch (JAXBException e) {
			throw new JAXRException(e);
		} catch (IOException e) {
			throw new JAXRException(e);
		}
	}

	/**
	 * Reads and deserializes a meta model from the XML stream provided
	 * by the given {@link XMLEventReader}.
     * @param pReader The StAX event reader, from which to read the meta model.
     * @return The created meta model.
     * @throws JAXRException The operation failed.
	 */
	public ROMetaModel read(XMLEventReader pReader) throws JAXRException {
		try {
			final net.sf.csutils.core.model.vo.ROMetaModel metaModel = jaxbReader.read(pReader);
			return read(metaModel);
		} catch (JAXBException e) {
			throw new JAXRException(e);
		}
	}

	/**
	 * Reads and deserializes a meta model from the XML stream provided
	 * by the given {@link XMLStreamReader}.
     * @param pReader The StAX stream reader, from which to read the meta model.
     * @return The created meta model.
     * @throws JAXRException The operation failed.
	 */
	public ROMetaModel read(XMLStreamReader pReader) throws JAXRException {
		try {
			final net.sf.csutils.core.model.vo.ROMetaModel metaModel = jaxbReader.read(pReader);
			return read(metaModel);
		} catch (JAXBException e) {
			throw new JAXRException(e);
		}
	}
}
