package net.sf.csutils.core.query;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import javax.xml.registry.JAXRException;

import net.sf.csutils.core.model.ROMetaModel;
import net.sf.csutils.core.query.impl.SqlQueryGenerator;
import net.sf.csutils.core.query.impl.XQueryGenerator;
import net.sf.csutils.core.registry.RegistryFacade;
import net.sf.csutils.core.registry.RegistryInfo;
import net.sf.csutils.core.registry.centrasite.CentraSiteRegistryInfo;
import net.sf.csutils.core.registry.jaxmas.JaxMasRegistryInfo;



/**
 * A parser for HQL statements. Converts HQL into XML query, suitable for
 * querying JAXR objects.
 */
public class CsqlParser {
    private static final String KEY_STATEMENT_CACHE = CsqlParser.class.getName() + ".statementCache";
	private final ROMetaModel model;
	private final RegistryInfo registryInfo;

	/**
	 * Creates a new instance with the given meta model, and registry info.
	 * @param pModel The registry object model to use.
     * @param pRegistryInfo The registry info object to use.
	 */
	public CsqlParser(ROMetaModel pModel, RegistryInfo pRegistryInfo) {
		model = pModel;
		registryInfo = pRegistryInfo;
	}

    /**
     * Creates a new instance with the given meta model, and the
     * {@link CentraSiteRegistryInfo#getInstance() CentraSite registry info}.
     * @param pModel The registry object model to use.
     */
	public CsqlParser(ROMetaModel pModel) {
	    this(pModel, CentraSiteRegistryInfo.getInstance());
	}

	@SuppressWarnings("unchecked")
	private Map<String,CsqlStatement> getStatementCache() throws JAXRException {
		Map<String, CsqlStatement> map = (Map<String, CsqlStatement>) model.getAttribute(KEY_STATEMENT_CACHE);
		if (map == null) {
			map = Collections.synchronizedMap(new HashMap<String,CsqlStatement>());
			model.setAttribute(KEY_STATEMENT_CACHE, map);
		}
		return map;
	}
	
	/**
	 * Parses the given query and converts it into an executable statement.
	 * @param pFacade The registry facade to use.
	 * @param pQuery The query string to parse.
	 * @return The parsed query.
	 * @throws JAXRException The operation failed.
	 */
	public CsqlStatement parse(RegistryFacade pFacade, String pQuery) throws JAXRException {
		Map<String,CsqlStatement> map = getStatementCache();
		CsqlStatement st = map.get(pQuery);
		if (st == null) {
		    if (registryInfo instanceof CentraSiteRegistryInfo) {
		        st = new XQueryGenerator().parse(model, new net.sf.csutils.core.query.tree.CsqlParser().parse(pQuery));
		    } else if (registryInfo instanceof JaxMasRegistryInfo) {
		        st = new SqlQueryGenerator().parse(model, new net.sf.csutils.core.query.tree.CsqlParser().parse(pQuery));
		    } else {
		        throw new IllegalStateException("Unknown type of registry info: " + registryInfo.getClass().getName());
		    }
			map.put(pQuery, st);
		}
		return st.clone(pFacade);
	}
}
