package net.sf.csutils.core.query.impl;

import java.util.ArrayList;
import java.util.Calendar;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.xml.registry.JAXRException;

import net.sf.csutils.core.model.QName;
import net.sf.csutils.core.model.ROAttribute;
import net.sf.csutils.core.model.ROMetaModel;
import net.sf.csutils.core.model.RORelation;
import net.sf.csutils.core.model.ROType;
import net.sf.csutils.core.model.impl.StaticROType;
import net.sf.csutils.core.query.CsqlStatement;
import net.sf.csutils.core.query.tree.AdditiveExpression;
import net.sf.csutils.core.query.tree.And;
import net.sf.csutils.core.query.tree.Concatenation;
import net.sf.csutils.core.query.tree.Eq;
import net.sf.csutils.core.query.tree.FunctionCall;
import net.sf.csutils.core.query.tree.Ge;
import net.sf.csutils.core.query.tree.Gt;
import net.sf.csutils.core.query.tree.Identifier;
import net.sf.csutils.core.query.tree.In;
import net.sf.csutils.core.query.tree.IsEmpty;
import net.sf.csutils.core.query.tree.IsNull;
import net.sf.csutils.core.query.tree.Le;
import net.sf.csutils.core.query.tree.Like;
import net.sf.csutils.core.query.tree.LogicalExpression;
import net.sf.csutils.core.query.tree.Lt;
import net.sf.csutils.core.query.tree.MethodCall;
import net.sf.csutils.core.query.tree.MultiplicativeExpression;
import net.sf.csutils.core.query.tree.Ne;
import net.sf.csutils.core.query.tree.Negation;
import net.sf.csutils.core.query.tree.ObjectType;
import net.sf.csutils.core.query.tree.Or;
import net.sf.csutils.core.query.tree.RelationalExpression;
import net.sf.csutils.core.query.tree.SelectStatement;
import net.sf.csutils.core.query.tree.Statement;
import net.sf.csutils.core.query.tree.UnaryMinus;


/**
 * Abstract base class for implementing query generators.
 */
public abstract class AbstractQueryGenerator {
    /**
     * The "query" namespace.
     */
    public static final String NS_QUERY = "http://namespaces.csutils.sf.net/query/1.0.0";
    /**
     * The "allTypes" QName.
     */
    public static final QName QNAME_ALLTYPES = new QName(NS_QUERY, "allTypes");

    public static class Parameter {
    	private final int offset;
    	private final int length;
    
    	public Parameter(int pOffset, int pLength) {
    		offset = pOffset;
    		length = pLength;
    	}
    
    	public int getOffset() {
    		return offset;
    	}
    
    	public int getLength() {
    		return length;
    	}
    }

    public static class NamedParameter extends Parameter {
    	private final String name;
    
    	public NamedParameter(String pName, int pOffset, int pLength) {
    		super(pOffset, pLength);
    		name = pName;
    	}
    
    	public String getName() {
    		return name;
    	}
    }

    public static class NumberedParameter extends Parameter {
    	private final int num;
    
    	public NumberedParameter(int pNum, int pOffset, int pLength) {
    		super(pOffset, pLength);
    		num = pNum;
    	}
    
    	public int getNum() {
    		return num;
    	}
    }

    protected static class ObjectTypeInfo {
        private final ROType roType;
        private final String alias;
        private final ObjectTypeInfo sourceType;
        private final RORelation relation;

        ObjectTypeInfo(String pAlias, ROType pROType, ObjectTypeInfo pSourceType, RORelation pRelation) {
            roType = pROType;
            alias = pAlias;
            sourceType = pSourceType;
            relation = pRelation;
        }

        public ROType getROType() {
            return roType;
        }

        public String getAlias() {
            return alias;
        }

        public ObjectTypeInfo getSourceType() {
            return sourceType;
        }

        public RORelation getRelation() {
            return relation;
        }
    }

    protected StringBuilder sb = new StringBuilder();
    private final Map<String,ObjectTypeInfo> objectTypes = new HashMap<String,ObjectTypeInfo>();
    private ROMetaModel model;
    protected int autoParameterNumber;
    protected final List<Parameter> parameters = new ArrayList<Parameter>();

    protected ROMetaModel getModel() {
        return model;
    }

    protected List<Parameter> getParameters() {
        return parameters;
    }

    protected void parse(And pAnd) throws JAXRException {
        final List<Object> items = pAnd.getItems();
        sb.append("(");
        for (int i = 0;  i < items.size();  i++) {
            if (i > 0) {
                sb.append(" and ");
            }
            parseExpression(items.get(i));
        }
        sb.append(")");
    }

    protected void parse(Or pOr) throws JAXRException {
        final List<Object> items = pOr.getItems();
        sb.append("(");
        for (int i = 0;  i < items.size();  i++) {
            if (i > 0) {
                sb.append(" or ");
            }
            parseExpression(items.get(i));
        }
        sb.append(")");
    }

    protected void parse(RelationalExpression pExpr, String pOp) throws JAXRException {
        parseExpression(pExpr.getItem1());
        sb.append(pOp);
        parseExpression(pExpr.getItem2());
    }

    protected abstract void parse(In pIn) throws JAXRException;
    
    protected void parse(Eq pEq) throws JAXRException {
        parse(pEq, "=");
    }

    protected void parse(Ne pNe) throws JAXRException {
        parse(pNe, "=");
    }

    protected void parse(Ge pGe) throws JAXRException {
        parse(pGe, ">=");
    }

    protected void parse(Gt pGt) throws JAXRException {
        parse(pGt, ">");
    }

    protected void parse(Le pLe) throws JAXRException {
        parse(pLe, "<=");
    }

    protected void parse(Lt pLt) throws JAXRException {
        parse(pLt, "<");
    }

    protected abstract void parse(Identifier pIdentifier) throws JAXRException;

    protected abstract void parse(MethodCall pMethodCall) throws JAXRException;

    protected abstract void parse(FunctionCall pFunctionCall) throws JAXRException;

    protected abstract void parse(Like pLike) throws JAXRException;

    protected abstract void parse(IsNull pIsNull) throws JAXRException;

    protected abstract void parse(Negation pNegation) throws JAXRException;

    protected abstract void parse(String pString) throws JAXRException;

    protected abstract void parse(Calendar pCalendar) throws JAXRException;

    protected abstract void parse(Boolean pBoolean) throws JAXRException;

    protected void parse(Concatenation pConcatenation) throws JAXRException {
        final List<Object> items = pConcatenation.getItems();
        sb.append("concat(");
        for (int i = 0;  i < items.size();  i++) {
            if (i > 0) {
                sb.append(",");
            }
            parseExpression(items.get(i));
        }
        sb.append(")");
    }

    protected void parse(AdditiveExpression pExpr) throws JAXRException {
        final List<Object> items = pExpr.getItems();
        for (int i = 0;  i < items.size();  i += 2) {
            if (i > 0) {
                final Object o = items.get(i-1);
                if (o == AdditiveExpression.PLUS) {
                    sb.append("+");
                } else {
                    sb.append("-");
                }
            }
            parseExpression(items.get(i));
        }
    }

    protected void parse(MultiplicativeExpression pExpr) throws JAXRException {
        final List<Object> items = pExpr.getItems();
        for (int i = 0;  i < items.size();  i += 2) {
            if (i > 0) {
                final Object o = items.get(i-1);
                if (o == MultiplicativeExpression.MULTIPLY) {
                    sb.append("*");
                } else {
                    sb.append("/");
                }
            }
            parseExpression(items.get(i));
        }
    }

    protected void parse(UnaryMinus pObject) throws JAXRException {
        sb.append("-");
        parseExpression(pObject.getItem());
    }

    /**
     * Checks, whether a string is empty.
     * @param pEmpty The empty value
     * @throws JAXRException Could not add the check for emptiness.
     */
    protected void parse(IsEmpty pEmpty) throws JAXRException {
        sb.append("''=");
        parseExpression(pEmpty.getItem());
    }

    /**
     * Adds the given number.
     * @param pValue The number to add to the query.
     * @throws JAXRException The number could not be added.
     */
    protected void parse(Number pValue) throws JAXRException {
        sb.append(pValue);
    }

    protected void parseExpression(Object pObject) throws JAXRException {
        if (pObject == null) {
            throw new IllegalArgumentException("The expression must not be null.");
        }
        if (pObject instanceof LogicalExpression) {
            if (pObject instanceof And) {
                parse((And) pObject);
                return;
            } else if (pObject instanceof Or) {
                parse((Or) pObject);
                return;
            }
        } else if (pObject instanceof RelationalExpression) {
            if (pObject instanceof Eq) {
                parse((Eq) pObject);
                return;
            } else if (pObject instanceof Ge) {
                parse((Ge) pObject);
                return;
            } else if (pObject instanceof Gt) {
                parse((Gt) pObject);
                return;
            } else if (pObject instanceof Le) {
                parse((Le) pObject);
                return;
            } else if (pObject instanceof Lt) {
                parse((Lt) pObject);
                return;
            } else if (pObject instanceof Ne) {
                parse((Ne) pObject);
                return;
            } else if (pObject instanceof Like) {
                parse((Like) pObject);
                return;
            }
        } else if (pObject instanceof Negation) {
            parse((Negation) pObject);
            return;
        } else if (pObject instanceof Concatenation) {
            parse((Concatenation) pObject);
            return;
        } else if (pObject instanceof AdditiveExpression) {
            parse((AdditiveExpression) pObject);
            return;
        } else if (pObject instanceof MultiplicativeExpression) {
            parse((MultiplicativeExpression) pObject);
            return;
        } else if (pObject instanceof UnaryMinus) {
            parse((UnaryMinus) pObject);
            return;
        } else if (pObject instanceof IsEmpty) {
            parse((IsEmpty) pObject);
            return;
        } else if (pObject instanceof Identifier) {
            parse((Identifier) pObject);
            return;
        } else if (pObject instanceof MethodCall) {
            parse((MethodCall) pObject);
            return;
        } else if (pObject instanceof FunctionCall) {
            parse((FunctionCall) pObject);
            return;
        } else if (pObject instanceof IsNull) {
            parse((IsNull) pObject);
            return;
        } else if (pObject instanceof net.sf.csutils.core.query.tree.Parameter) {
            parse((net.sf.csutils.core.query.tree.Parameter) pObject);
            return;
        } else if (pObject instanceof String) {
            parse((String) pObject);
            return;
        } else if (pObject instanceof Calendar) {
            parse((Calendar) pObject);
            return;
        } else if (pObject instanceof Number) {
            parse((Number) pObject);
            return;
        } else if (pObject instanceof Boolean) {
            parse((Boolean) pObject);
            return;
        } else if (pObject instanceof In) {
        	parse((In) pObject);
        	return;
        }
        throw new IllegalStateException("Invalid expression type: " + pObject.getClass().getName());
    }

    /**
     * Adds the given parameter to the query.
     * @param pParameter The parameter to add to the query.
     * @throws JAXRException Failed to add the parameter.
     */
    protected void parse(net.sf.csutils.core.query.tree.Parameter pParameter) throws JAXRException {
        final String name = pParameter.getName();
        if (name == null) {
            final Integer number = pParameter.getNumber();
            final int num;
            if (number == null) {
                num = autoParameterNumber++;
            } else {
                num = number.intValue();
            }
            final String param = String.valueOf(num);
            int offset = sb.length();
            int length = param.length() + 1;
            sb.append("?");
            sb.append(param);
            parameters.add(new NumberedParameter(num, offset, length));
        } else {
            int offset = sb.length();
            int length = name.length() + 1;
            sb.append(":");
            sb.append(name);
            parameters.add(new NamedParameter(name, offset, length));
        }
    }

    protected ObjectTypeInfo getAlias(String pPath, String pAlias) throws JAXRException {
        final ObjectTypeInfo oti = getObjectType(pAlias);
        if (oti == null) {
            throw new JAXRException("Invalid identifier " + pPath
                    + ": The alias " + pAlias + " is unknown.");
        }
        return oti;
    }

    private ROType getAllTypes(final QName qName) {
        final StaticROType otype = new StaticROType();
        otype.setQName(qName);
        otype.setAttributes(new HashMap<String,ROAttribute>());
        return otype;
    }

    protected ObjectTypeInfo createJoin(ObjectType pQueryType) throws JAXRException {
        final String identifier = pQueryType.getName();
        final int offset = identifier.indexOf('.');
        if (offset <= 0  ||  offset > identifier.length()-2) {
            throw new JAXRException("Unable to parse join identifier: Expected <alias>.<relation>, got "
                    + identifier);
        }
        final String alias = identifier.substring(0, offset);
        final String relation = identifier.substring(offset+1);
        final ObjectTypeInfo sourceType = objectTypes.get(alias);
        if (alias == null) {
            throw new JAXRException("Invalid join identifier " + identifier
                    + ": The alias " + alias + " is unknown.");
        }
        if (relation.indexOf('.') != -1) {
            throw new JAXRException("Invalid join identifier " + identifier
                    + ": Only one join level is currently supported.");
        }
        final ROAttribute roAttribute = sourceType.getROType().getAttribute(relation);
        if (roAttribute == null) {
            throw new JAXRException("Invalid join identifier " + identifier
                    + ": The object type " + sourceType.getROType().getQName()
                    + " has no attribute " + relation);
        }
        if (roAttribute.getType() != ROAttribute.Type.relation) {
            throw new JAXRException("Invalid join identifier " + identifier
                    + ": object type " + sourceType.getROType().getQName()
                    + " has an attribute " + relation
                    + ", but the attribute type is not 'relation'.");
        }
        final RORelation roRelation = (RORelation) roAttribute;
        final List<ROType> targetTypes = roRelation.getTargetTypes();
        if (targetTypes.size() > 1) {
            throw new JAXRException("Relations with multiple target types are currently unsupported.");
        }
        return newObjectTypeInfo(pQueryType.getAlias(), targetTypes.get(0), sourceType, roRelation);
    }

    /**
     * Creates a new instance of {@link ObjectTypeInfo}. This is invoked for the
     * main registry object type in the query as well as for any JOIN clause.
     * @param pAlias The alias to use for the registry object type.
     * @param pType The registry object type.
     * @param pSourceType Contains the left hand side of the join clause
     *   (the source type of a relationship attribute), if the method is
     *   invoked to implement such a clause, otherwise null.
     * @param pAttribute Contains the relationship attribute of the
     *   join clause, if the method is invoked to implement such a clause,
     *   otherwise null.
     * @return The created object.
     * @throws JAXRException Creating the ObjectTypeInfo failed. 
     */
    protected ObjectTypeInfo newObjectTypeInfo(String pAlias, ROType pType, ObjectTypeInfo pSourceType, RORelation pAttribute) throws JAXRException {
        return new ObjectTypeInfo(pAlias, pType, pSourceType, pAttribute);
    }
    
    protected ROType findMatchingObjectType(Statement pStatement, ObjectType pQueryType) throws JAXRException {
        final String name = pQueryType.getName();
        int offset = name.indexOf(':');
        if (offset == -1) {
            if ("*".equals(name)) {
                return getAllTypes(QNAME_ALLTYPES);
            }
            ROType otype = null;
            for (ROType type : getModel().getROTypes().values()) {
                if (type.getQName().getLocalPart().equals(pQueryType.getName())) {
                    if (otype == null) {
                        otype = type;
                    } else {
                        throw new JAXRException("Multiple matching object types found for "
                                + pQueryType.getName() + ": "
                                + otype.getQName() + ", " + type.getQName());
                    }
                }
            }
            if (otype == null) {
                throw new JAXRException("No matching object type found for " + pQueryType.getName());
            }
            return otype;
        } else {
            final String prefix = name.substring(0, offset);
            final String localPart = name.substring(offset+1);
            final String uri = pStatement.getNamespaceUri(prefix);
            if (uri == null) {
                throw new JAXRException("Undeclared namespace prefix: " + name);
            }
            final QName qName = new QName(uri, localPart, prefix);
            if (NS_QUERY.equals(uri)) {
                if ("allTypes".equals(localPart)) {
                    return getAllTypes(qName);
                } else {
                    throw new JAXRException("Unknown object type: " + qName
                            + ", did you mean " + QNAME_ALLTYPES + "?");
                }
            } else {
                final ROType otype = getModel().getROType(qName);
                if (otype == null) {
                    throw new JAXRException("Unknown object type: " + qName);
                }
                return otype;
            }
        }
    }

    protected ObjectTypeInfo getObjectType(String pAlias) {
        return objectTypes.get(pAlias);
    }

    protected List<ObjectTypeInfo> findObjectTypes(SelectStatement pQuery) throws JAXRException {
        final List<ObjectTypeInfo> otypes = new ArrayList<ObjectTypeInfo>();
        final List<ObjectType> ots = pQuery.getObjectTypes();
        boolean first = true;
        for (ObjectType ot : ots) {
            final ObjectTypeInfo objectTypeInfo;
            final ROType otype;
            if (first) {
                otype = findMatchingObjectType(pQuery, ot);
                objectTypeInfo = newObjectTypeInfo(ot.getAlias(), otype, null, null);
                first = false;
            } else {
                objectTypeInfo = createJoin(ot);
                otype = objectTypeInfo.getROType();
            }
            objectTypes.put(ot.getAlias(), objectTypeInfo);
            otypes.add(objectTypeInfo);
        }
        return otypes;
    }

    protected abstract CsqlStatement parse(SelectStatement pQuery) throws JAXRException;
    
    public CsqlStatement parse(ROMetaModel pModel, Statement pQuery) throws JAXRException {
        if (pModel == null) {
            throw new IllegalArgumentException("The meta model must not be null.");
        }
        model = pModel;
        if (pQuery == null) {
            throw new IllegalArgumentException("The query must not be null.");
        }
        if (pQuery instanceof SelectStatement) {
            return parse((SelectStatement) pQuery);
        }
        throw new IllegalStateException("Invalid statement type: " + pQuery.getClass().getName());
    }
}
