package net.sf.csutils.core.query.impl;

import java.sql.SQLException;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collections;
import java.util.List;

import javax.xml.namespace.QName;
import javax.xml.registry.JAXRException;

import net.sf.csutils.core.model.ROAttribute;
import net.sf.csutils.core.model.RORelation;
import net.sf.csutils.core.model.ROType;
import net.sf.csutils.core.query.CsqlStatement;
import net.sf.csutils.core.query.tree.And;
import net.sf.csutils.core.query.tree.FunctionCall;
import net.sf.csutils.core.query.tree.Identifier;
import net.sf.csutils.core.query.tree.In;
import net.sf.csutils.core.query.tree.IsNull;
import net.sf.csutils.core.query.tree.Like;
import net.sf.csutils.core.query.tree.MethodCall;
import net.sf.csutils.core.query.tree.Negation;
import net.sf.csutils.core.query.tree.Or;
import net.sf.csutils.core.query.tree.RelationalExpression;
import net.sf.csutils.core.query.tree.SelectStatement;
import net.sf.csutils.core.registry.centrasite.CentraSiteRegistryInfo;

import org.apache.labs.jaxmas.registry.infomodel.InternationalStringImpl;
import org.apache.labs.jaxmas.registry.sql.FilterPredicate;
import org.apache.labs.jaxmas.registry.sql.RegistryObjectLoader;
import org.apache.labs.jaxmas.registry.sql.RegistryObjectLoader.RegistryObjectQuery;



/**
 * Converts a CSQL query into XQuery.
 */
public class SqlQueryGenerator extends AbstractQueryGenerator {
    private abstract class SubQueryHandler {
        void createSubQuery(int pOffset) throws JAXRException {
            final String value = sb.substring(pOffset);
            sb.setLength(pOffset);
            createSubQuery(value);
        }

        abstract void createSubQuery(String pValue) throws JAXRException;
    }

    private static class JoinPredicate implements FilterPredicate {
        private final RegistryObjectQuery source, target;
        private final String attributeKey;

        JoinPredicate(RegistryObjectQuery pSource, RegistryObjectQuery pTarget, String pAttributeKey) {
            source = pSource;
            target = pTarget;
            attributeKey = pAttributeKey;
        }

        @Override
        public void add(StringBuilder pSb, List<Object> pParameters)
                throws SQLException, JAXRException {
            final String asSource = RegistryObjectLoader.getAliasAssociations(source);
            final String asRoSource = RegistryObjectLoader.getAliasAssociations(source) + "ro";
            final String roSource = RegistryObjectLoader.getAliasRegistryObjects(source);
            final String rosSource = RegistryObjectLoader.getAliasSlots(source);
            final String rosvSource = RegistryObjectLoader.getAliasSlotValues(source);
            final String roTarget = RegistryObjectLoader.getAliasRegistryObjects(target);
            pSb.append("EXISTS(SELECT ");
            pSb.append(asSource);
            pSb.append(".roKey FROM RegistryObjects ");
            pSb.append(asRoSource);
            pSb.append(" JOIN Associations ");
            pSb.append(asSource);
            pSb.append(" ON ");
            pSb.append(asRoSource);
            pSb.append(".roKey=");
            pSb.append(asSource);
            pSb.append(".roKey AND ");
            pSb.append(asRoSource);
            pSb.append(".roOwnerCascading=");
            pSb.append(roSource);
            pSb.append(".roKey AND ");
            pSb.append(asSource);
            pSb.append(".roKeyTarget=");
            pSb.append(roTarget);
            pSb.append(".roKey JOIN RegistryObjectSlots ");
            pSb.append(rosSource);
            pSb.append(" ON ");
            pSb.append(asRoSource);
            pSb.append(".roKey=");
            pSb.append(rosSource);
            pSb.append(".roKey JOIN RegistryObjectSlotValues ");
            pSb.append(rosvSource);
            pSb.append(" ON ");
            pSb.append(rosSource);
            pSb.append(".id=");
            pSb.append(rosvSource);
            pSb.append(".id AND ");
            pSb.append(rosSource);
            pSb.append(".name=? AND ");
            pParameters.add(CentraSiteRegistryInfo.SLOT_NAME_ATTRIBUTES_KEY);
            pSb.append(rosvSource);
            pSb.append(".val=?");
            pParameters.add(attributeKey);
            pSb.append(")");
        }
    }
    
    private static class MyObjectTypeInfo extends ObjectTypeInfo {
        RegistryObjectQuery roq;
        MyObjectTypeInfo(String pAlias, ROType pROType,
                ObjectTypeInfo pSourceType, RORelation pRelation) throws JAXRException {
            super(pAlias, pROType, pSourceType, pRelation);
            if (QNAME_ALLTYPES.equals(pROType.getQName())) {
                roq = RegistryObjectLoader.getRegistryObjectQuery(null, pAlias);
            } else {
                roq = RegistryObjectLoader.getRegistryObjectQuery(pROType.getQName().toString(), pAlias);
            }
            if (pRelation != null) {
                final MyObjectTypeInfo source = (MyObjectTypeInfo) pSourceType;
                final List<FilterPredicate> predicates = new ArrayList<FilterPredicate>();
                if (roq.getPredicates() != null) {
                    predicates.addAll(roq.getPredicates());
                }
                predicates.add(new JoinPredicate(source.roq, roq, pRelation.getAttributeKey()));
                roq = new RegistryObjectQuery(roq.getObjectType(), pAlias, predicates);
            }
        }
    }

    private static final DateFormat dateTimeFormat = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss.SSSSSSS");
    private final List<SubQueryHandler> subQueryHandlers = new ArrayList<SubQueryHandler>();

    private List<Integer> subQueryOffsets = new ArrayList<Integer>();
    private void createSubqueryContext() {
        subQueryOffsets.add(Integer.valueOf(sb.length()));
    }

    private void applySubqueries() throws JAXRException {
        final int offset = subQueryOffsets.remove(subQueryOffsets.size()-1).intValue();
        for (SubQueryHandler subQuery : subQueryHandlers) {
            subQuery.createSubQuery(offset);
        }
        subQueryHandlers.clear();
    }
    
    @Override
    protected ObjectTypeInfo newObjectTypeInfo(String pAlias, ROType pType, ObjectTypeInfo pSourceType, RORelation pAttribute)
            throws JAXRException {
        return new MyObjectTypeInfo(pAlias, pType, pSourceType, pAttribute);
    }
    
    /**
     * @throws JAXRException Failed to add the parameter.
     */
    @Override
    protected void parse(net.sf.csutils.core.query.tree.Parameter pParameter) throws JAXRException {
        final String name = pParameter.getName();
        if (name == null) {
            final Integer number = pParameter.getNumber();
            final int num;
            if (number == null) {
                num = autoParameterNumber++;
            } else {
                num = number.intValue();
            }
            final String param = String.valueOf(num);
            int offset = sb.length();
            int length = param.length() + 1;
            sb.append("?");
            parameters.add(new NumberedParameter(num, offset, length));
        } else {
            int offset = sb.length();
            int length = name.length() + 1;
            sb.append(":");
            sb.append(name);
            parameters.add(new NamedParameter(name, offset, length));
        }
    }

    @Override
    protected void parse(Like pLike) throws JAXRException {
        parseExpression(pLike.getItem1());
        sb.append(" LIKE ");
        parseExpression(pLike.getItem2());
    }

    @Override
    protected void parse(In pIn) throws JAXRException {
    	parseExpression(pIn.getLhs());
    	sb.append(" IN (");
    	final List<?> rhs = pIn.getRhs();
    	for (int i = 0;  i < rhs.size();  i++) {
    		if (i > 0) {
    			sb.append(", ");
    		}
    		parseExpression(rhs.get(i));
    	}
    	sb.append(")");
    }
    
    
    @Override
    protected void parse(And pAnd) throws JAXRException {
        createSubqueryContext();
        super.parse(pAnd);
        applySubqueries();
    }

    @Override
    protected void parse(Or pOr) throws JAXRException {
        createSubqueryContext();
        super.parse(pOr);
        applySubqueries();
    }

    @Override
    protected void parse(RelationalExpression pExpr, String pOp) throws JAXRException {
        createSubqueryContext();
        super.parse(pExpr, pOp);
        applySubqueries();
    }

    private void checkNoArgs(String pMethodName, List<Object> pArgs) throws JAXRException {
        if (!pArgs.isEmpty()) {
            throw new JAXRException("The method " + pMethodName + " allows no arguments.");
        }
    }

    private void checkZeroOrOneArgs(String pMethodName, List<Object> pArgs) throws JAXRException {
        if (pArgs.size() > 1) {
            throw new JAXRException("The method " + pMethodName + " must have 0 or 1 argument.");
        }
    }

    protected ObjectTypeInfo parseRoAlias(String pPath, String pAlias) throws JAXRException {
        final ObjectTypeInfo oti = getObjectType(pAlias);
        if (pAlias == null) {
            throw new JAXRException("Invalid identifier " + pPath
                    + ": The alias " + pAlias + " is unknown.");
        }
        sb.append(oti.getAlias());
        return oti;
    }

    protected ObjectTypeInfo parseAlias(String pPath, String pAlias) throws JAXRException {
        final ObjectTypeInfo oti = getAlias(pPath, pAlias);
        sb.append(oti.getAlias());
        return oti;
    }

    private int numNameOrDescriptionQueries;
    
    protected void parseNameOrDescription(final InternationalStringImpl.Type pType, String pPath, final List<Object> pArgs) throws JAXRException {
        final MyObjectTypeInfo oti = (MyObjectTypeInfo) getAlias(pPath, pPath);
        final String roTable = RegistryObjectLoader.getAliasRegistryObjects(oti.roq);
        final String alias = oti.getAlias() + "ls" + numNameOrDescriptionQueries++;
        subQueryHandlers.add(new SubQueryHandler() {
            @Override
            void createSubQuery(String pValue) throws JAXRException {
                sb.append("EXISTS(SELECT ");
                sb.append(alias);
                sb.append(".roKey FROM LocalizedStrings ");
                sb.append(alias);
                sb.append(" WHERE ");
                sb.append(alias);
                sb.append(".roKey=");
                sb.append(roTable);
                sb.append(".roKey AND ");
                sb.append(alias);
                sb.append(".lsType=");
                sb.append(pType.ordinal());
                if (!pArgs.isEmpty()) {
                    sb.append(" AND ");
                    sb.append(alias);
                    sb.append(".locale=");
                    parseExpression(pArgs.get(0));
                }
                sb.append(" AND ");
                sb.append(pValue);
                sb.append(')');
            }
        });
        sb.append(alias);
        sb.append(".val");
    }
    
    @Override
    protected void parse(MethodCall pMethodCall) throws JAXRException {
        final String methodName = pMethodCall.getMethodName();
        final String path = pMethodCall.getItem().getPath();
        final List<Object> args = pMethodCall.getArgs();
        int offset = path.indexOf('.');
        if (offset != -1) {
            throw new JAXRException("Invalid method call: Expected <alias>."
                    + methodName + "(...), got " + path);
        }
        if ("key".equals(methodName)) {
            checkNoArgs(methodName, args);
            final MyObjectTypeInfo oti = (MyObjectTypeInfo) getObjectType(path);
            final String table = RegistryObjectLoader.getAliasRegistryObjects(oti.roq);
            sb.append(table);
            sb.append(".roKey");
        } else if ("status".equals(methodName)) {
            checkNoArgs(methodName, args);
            final MyObjectTypeInfo oti = (MyObjectTypeInfo) getObjectType(path);
            final String table = RegistryObjectLoader.getAliasRegistryEntries(oti.roq);
            sb.append(table);
            sb.append(".status");
        } else if ("stability".equals(methodName)) {
            checkNoArgs(methodName, args);
            final MyObjectTypeInfo oti = (MyObjectTypeInfo) getObjectType(path);
            final String table = RegistryObjectLoader.getAliasRegistryEntries(oti.roq);
            sb.append(table);
            sb.append(".stability");
        } else if ("minorVersion".equals(methodName)) {
            checkNoArgs(methodName, args);
            final MyObjectTypeInfo oti = (MyObjectTypeInfo) getObjectType(path);
            final String table = RegistryObjectLoader.getAliasRegistryEntries(oti.roq);
            sb.append(table);
            sb.append(".minorVersion");
        } else if ("majorVersion".equals(methodName)) {
            checkNoArgs(methodName, args);
            final MyObjectTypeInfo oti = (MyObjectTypeInfo) getObjectType(path);
            final String table = RegistryObjectLoader.getAliasRegistryEntries(oti.roq);
            sb.append(table);
            sb.append(".majorVersion");
        } else if ("userVersion".equals(methodName)) {
            checkNoArgs(methodName, args);
            final MyObjectTypeInfo oti = (MyObjectTypeInfo) getObjectType(path);
            final String table = RegistryObjectLoader.getAliasRegistryEntries(oti.roq);
            sb.append(table);
            sb.append(".userVersion");
        } else if ("name".equals(methodName)) {
            checkZeroOrOneArgs(methodName, args);
            parseNameOrDescription(InternationalStringImpl.Type.name, path, args);
        } else if ("description".equals(methodName)) {
            checkZeroOrOneArgs(methodName, args);
            parseNameOrDescription(InternationalStringImpl.Type.name, path, args);
        } else {
            throw new JAXRException("Unknown method name: " + methodName);
        }
    }

    @Override
    protected CsqlStatement parse(SelectStatement pQuery) throws JAXRException {
        final List<ObjectTypeInfo> otypes = findObjectTypes(pQuery);
        final RegistryObjectQuery[] roqs = new RegistryObjectQuery[otypes.size()];
        for (int i = 0;  i < otypes.size();  i++) {
            RegistryObjectQuery roq = ((MyObjectTypeInfo) otypes.get(i)).roq;
            if (i == 0) {
                final Object o = pQuery.getWhereClause();
                if (o != null) {
                    createSubqueryContext();
                    parseExpression(pQuery.getWhereClause());
                    applySubqueries();
                    final FilterPredicate filterPredicate = new FilterPredicate(){
                        @Override
                        public void add(StringBuilder pBuffer,
                                List<Object> pParameters) throws SQLException,
                                JAXRException {
                            pBuffer.append(sb);
                        }
                    };
                    final List<FilterPredicate> filterPredicates;
                    if (roq.getPredicates() == null  ||  roq.getPredicates().isEmpty()) {
                        filterPredicates = Collections.singletonList(filterPredicate);
                    } else {
                        filterPredicates = new ArrayList<FilterPredicate>();
                        filterPredicates.addAll(roq.getPredicates());
                        filterPredicates.add(filterPredicate);
                    }
                    roq = new RegistryObjectQuery(roq.getObjectType(), roq.getPrefix(), filterPredicates);
                }
            }
            roqs[i] = roq;
        }
        try {
            final RegistryObjectLoader rol = new RegistryObjectLoader(roqs, null);
            return new SqlCsqlStatement(null, rol);
        } catch (SQLException e) {
            throw new JAXRException(e);
        }
    }

    @Override
    protected void parse(Identifier pIdentifier) throws JAXRException {
        final String path = pIdentifier.getPath();
        final int offset = path.indexOf('.');
        if (offset <= 0  ||  offset > path.length()-2  ||  path.indexOf('.', offset+1) != -1) {
            throw new JAXRException("Invalid identifier: Expected <alias>.<attribute>, got "
                    + path);
        }
        final String alias = path.substring(0, offset);
        final String attribute = path.substring(offset+1);
        final ObjectTypeInfo oti = getAlias(path, alias);
        final ROAttribute roAttribute = oti.getROType().getAttribute(attribute);
        if (roAttribute == null) {
            throw new JAXRException("Invalid identifier " + path
                    + ": The object type " + oti.getROType().getQName()
                    + " has no attribute " + attribute);
        }
        switch (roAttribute.getType()) {
            case string:
            case bool:
            case date:
            case dateTime:
            case emailAddress:
            case integer:
            case ipv4Address:
            case ipv6Address:
            case longInteger:
            case number:
            case time:
            case uri:
                break;
            default:
                throw new JAXRException("Invalid identifier " + path
                        + ": attribute type of " + attribute
                        + " in object type " + oti.getROType().getQName()
                        + " is " + roAttribute.getType() + ", which is unsupported in expressions.");
        }
    	final RegistryObjectQuery roq = ((MyObjectTypeInfo) oti).roq;
        final String ro = RegistryObjectLoader.getAliasRegistryObjects(roq);
        final String ros = RegistryObjectLoader.getAliasSlots(roq);
        final String rosv = RegistryObjectLoader.getAliasSlotValues(roq);
        subQueryHandlers.add(new SubQueryHandler() {
            @Override
            void createSubQuery(String pValue) throws JAXRException {
                sb.append("EXISTS(SELECT ");
                sb.append(ros);
                sb.append(".roKey FROM RegistryObjectSlots ");
                sb.append(ros);
                sb.append(" JOIN RegistryObjectSlotValues ");
                sb.append(rosv);
                sb.append(" ON ");
                sb.append(ros);
                sb.append(".id=");
                sb.append(rosv);
                sb.append(".id WHERE ");
                sb.append(ros);
                sb.append(".roKey=");
                sb.append(ro);
                sb.append(".roKey AND ");
                sb.append(ros);
                sb.append(".name=");
                parse(new QName(oti.getROType().getQName().getNamespaceURI(), attribute).toString());
                sb.append(" AND ");
                sb.append(pValue);
                sb.append(')');
            }
        });
        sb.append(rosv);
        sb.append(".val");
    }

    @Override
    protected void parse(FunctionCall pFunctionCall) throws JAXRException {
        final String functionName = pFunctionCall.getFunctionName().toLowerCase();
        final String xQueryFunctionName;
        final int numArgs;
        if ("upper".equals(functionName)) {
            xQueryFunctionName = "UPPER";
            numArgs = 1;
        } else if ("lower".equals(functionName)) {
            xQueryFunctionName = "LOWER";
            numArgs = 1;
        } else {
            throw new JAXRException("Invalid function name: " + functionName);
        }
        final List<Object> args = pFunctionCall.getArgs();
        if (numArgs != args.size()) {
            throw new JAXRException("Expected exactly " + numArgs
                    + " arguments for function " + functionName
                    + ", got " + args.size());
        }
        sb.append(xQueryFunctionName);
        if (args.isEmpty()) {
            sb.append("()");
        } else {
            for (int i = 0;  i < args.size();  i++) {
                sb.append(i == 0 ? '(' : ',');
                parseExpression(args.get(i));
            }
            sb.append(')');
        }
    }

    @Override
    protected void parse(IsNull pIsNull) throws JAXRException {
        parseExpression(pIsNull.getItem());
        sb.append(" IS NULL");
    }

    @Override
    protected void parse(Negation pNegation) throws JAXRException {
        sb.append("NOT ");
        parseExpression(pNegation.getItem());
    }

    @Override
    protected void parse(String pString) throws JAXRException {
        sb.append("'");
        sb.append(pString);
        sb.append("'");
    }

    @Override
    protected void parse(Calendar pCalendar) throws JAXRException {
        final DateFormat format = (DateFormat) dateTimeFormat.clone();
        sb.append("'");
        sb.append(format.format(pCalendar.getTime()));
        sb.append("'");
    }

    @Override
    protected void parse(Boolean pBoolean) throws JAXRException {
        if (pBoolean.booleanValue()) {
            sb.append("true");
        } else {
            sb.append("false");
        }
    }
}
