package net.sf.csutils.core.query.impl;

import static net.sf.csutils.core.utils.Generics.cast;

import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.xml.namespace.QName;
import javax.xml.registry.BulkResponse;
import javax.xml.registry.BusinessLifeCycleManager;
import javax.xml.registry.BusinessQueryManager;
import javax.xml.registry.JAXRException;
import javax.xml.registry.infomodel.Key;
import javax.xml.registry.infomodel.RegistryObject;
import javax.xml.xquery.XQException;
import javax.xml.xquery.XQExpression;
import javax.xml.xquery.XQResultSequence;

import net.sf.csutils.core.query.CsqlStatement;
import net.sf.csutils.core.query.impl.AbstractQueryGenerator.NamedParameter;
import net.sf.csutils.core.query.impl.AbstractQueryGenerator.NumberedParameter;
import net.sf.csutils.core.query.impl.AbstractQueryGenerator.Parameter;
import net.sf.csutils.core.registry.RegistryFacade;
import net.sf.csutils.core.registry.centrasite.CentraSiteMetaModelAccessor;

import org.apache.log4j.Logger;

import com.centrasite.jaxr.CentraSiteConnection;


public class XQueryCsqlStatement implements CsqlStatement {
    private static final Map<String,String> BUILTIN_NAME_MAP = new HashMap<String,String>();
    static {
        final Map<net.sf.csutils.core.model.QName,net.sf.csutils.core.model.QName> map = CentraSiteMetaModelAccessor.BUILTIN_NAME_MAP;
        for (Map.Entry<net.sf.csutils.core.model.QName,net.sf.csutils.core.model.QName> entry : map.entrySet()) {
            BUILTIN_NAME_MAP.put(entry.getValue().toString(), entry.getKey().toString());
        }
    }
	private static class Date {
		private Calendar cal;

		Date(Calendar pCal) {
			cal = pCal;
		}

		@SuppressWarnings("unused")
        Calendar getCalendar() {
			return cal;
		}
	}

	private static class Time {
		private Calendar cal;

		Time(Calendar pCal) {
			cal = pCal;
		}

		Calendar getCalendar() {
			return cal;
		}
	}

	private static final Logger log = Logger.getLogger(XQueryCsqlStatement.class);
	private final String xQuery;
	private final List<net.sf.csutils.core.model.QName> objectTypes;
	private final List<Parameter> parametersByOffset;
	private final Set<String> namedParameterNames;
	private final List<Integer> numberedParameters;
	private final Map<String,Object> namedParameterValues;
	private final List<Object> numberedParameterValues;
	private final int parameterOffset;
	private final RegistryFacade facade;

	private static final Comparator<Parameter> comparator = new Comparator<Parameter>(){
		@Override
		public int compare(Parameter pParam1, Parameter pParam2) {
			return pParam1.getOffset() - pParam2.getOffset();
		}
	};

	private static List<Parameter> sort(List<Parameter> pParameters) {
		Collections.sort(pParameters, comparator);
		return pParameters;
	}

	private static final Set<String> parameterNames(List<Parameter> pParameters) {
		final Set<String> names = new HashSet<String>();
		for (Parameter p : pParameters) {
			if (p instanceof NamedParameter) {
				names.add(((NamedParameter) p).getName());
			}
		}
		return names;
	}

	private static final List<Integer> parameterNumbers(List<Parameter> pParameters) {
		final List<Integer> numbers = new ArrayList<Integer>();
		for (Parameter p : pParameters) {
			if (p instanceof NumberedParameter) {
				final int num = ((NumberedParameter) p).getNum();
				numbers.add(num, Integer.valueOf(num));
			}
		}
		return numbers;
	}
	
	public XQueryCsqlStatement(String pXQuery, List<net.sf.csutils.core.model.QName> pObjectTypes, List<Parameter> pParameters, int pParameterOffset) {
		this(null, pXQuery, pObjectTypes, sort(pParameters), parameterNames(pParameters), parameterNumbers(pParameters), pParameterOffset);
	}

	private XQueryCsqlStatement(RegistryFacade pFacade, String pXQuery, List<net.sf.csutils.core.model.QName> pObjectTypes, List<Parameter> pParameters, Set<String> pParameterNames, List<Integer> pParameterNumbers, int pParameterOffset) {
	    facade = pFacade;
	    xQuery = pXQuery;
		objectTypes = pObjectTypes;
		parametersByOffset = pParameters;
		namedParameterNames = pParameterNames;
		numberedParameters = pParameterNumbers;
		namedParameterValues = new HashMap<String,Object>();
        parameterOffset = pParameterOffset;
		numberedParameterValues = new ArrayList<Object>(numberedParameters.size());
		for (int i = 0;  i < numberedParameters.size();  i++) {
		    numberedParameterValues.add(null);
		}
	}

	/**
	 * Creates a cloned instance of this {@link XQueryCsqlStatement}.
	 */
	@Override
	public XQueryCsqlStatement clone(RegistryFacade pFacade) {
		return new XQueryCsqlStatement(pFacade, xQuery, objectTypes, parametersByOffset, namedParameterNames, numberedParameters, parameterOffset);
	}

	@Override
    public void setString(int pNum, String pValue) {
		numberedParameterValues.set(pNum, pValue);
	}

    @Override
	public void setString(String pName, String pValue) {
		namedParameterValues.put(pName, pValue);
	}

    @Override
	public void setByte(int pNum, byte pValue) {
		numberedParameterValues.set(pNum, Byte.valueOf(pValue));
	}

    @Override
	public void setByte(String pName, byte pValue) {
		namedParameterValues.put(pName, Byte.valueOf(pValue));
	}

    @Override
	public void setShort(int pNum, short pValue) {
		numberedParameterValues.set(pNum, Short.valueOf(pValue));
	}

    @Override
	public void setShort(String pName, short pValue) {
		namedParameterValues.put(pName, Short.valueOf(pValue));
	}

    @Override
	public void setInt(int pNum, int pValue) {
		numberedParameterValues.set(pNum, Integer.valueOf(pValue));
	}

    @Override
	public void setInt(String pName, int pValue) {
		namedParameterValues.put(pName, Integer.valueOf(pValue));
	}

    @Override
	public void setLong(int pNum, long pValue) {
		numberedParameterValues.set(pNum, Long.valueOf(pValue));
	}

    @Override
	public void setLong(String pName, long pValue) {
		namedParameterValues.put(pName, Long.valueOf(pValue));
	}

    @Override
	public void setFloat(int pNum, float pValue) {
		numberedParameterValues.set(pNum, Float.valueOf(pValue));
	}

    @Override
	public void setFloat(String pName, float pValue) {
		namedParameterValues.put(pName, Float.valueOf(pValue));
	}

    @Override
	public void setDouble(int pNum, double pValue) {
		numberedParameterValues.set(pNum, Double.valueOf(pValue));
	}

    @Override
	public void setDouble(String pName, double pValue) {
		namedParameterValues.put(pName, Double.valueOf(pValue));
	}

    @Override
	public void setDateTime(int pNum, Calendar pCalendar) {
		numberedParameterValues.set(pNum, pCalendar);
	}

    @Override
	public void setDateTime(String pName, Calendar pCalendar) {
		namedParameterValues.put(pName, pCalendar);
	}

    @Override
	public void setDate(int pNum, Calendar pCalendar) {
		numberedParameterValues.set(pNum, new Date(pCalendar));
	}

    @Override
	public void setDate(String pName, Calendar pCalendar) {
		namedParameterValues.put(pName, new Date(pCalendar));
	}

    @Override
	public void setTime(int pNum, Calendar pCalendar) {
		numberedParameterValues.set(pNum, new Time(pCalendar));
	}

    @Override
	public void setTime(String pName, Calendar pCalendar) {
		namedParameterValues.put(pName, new Time(pCalendar));
	}

    @Override
    public void setBoolean(int pNum, boolean pValue) {
        numberedParameterValues.set(pNum, Boolean.valueOf(pValue));
    }

    @Override
    public void setBoolean(String pName, boolean pValue) {
	    namedParameterValues.put(pName, Boolean.valueOf(pValue));
	}
	
	public String getXQuery() {
		return xQuery;
	}

	protected String bindParam(XQExpression pExpression, String pVar, Object pObject) throws XQException, JAXRException {
	    if (pObject instanceof Boolean) {
	        pExpression.bindBoolean(new QName(pVar), ((Boolean) pObject).booleanValue());
	    }
		if (pObject instanceof String) {
			pExpression.bindString(new QName(pVar), (String) pObject);
			return "xs:string";
		}
		if (pObject instanceof Byte) {
			pExpression.bindByte(new QName(pVar), ((Byte) pObject).byteValue());
			return "xs:byte";
		}
		if (pObject instanceof Short) {
			pExpression.bindInt(new QName(pVar), ((Short) pObject).intValue());
			return "xs:int";
		}
		if (pObject instanceof Integer) {
			pExpression.bindInt(new QName(pVar), ((Integer) pObject).intValue());
			return "xs:int";
		}
		if (pObject instanceof Long) {
			pExpression.bindLong(new QName(pVar), ((Long) pObject).longValue());
			return "xs:long";
		}
		if (pObject instanceof Float) {
			pExpression.bindFloat(new QName(pVar), ((Float) pObject).floatValue());
			return "xs:float";
		}
		if (pObject instanceof Double) {
			pExpression.bindDouble(new QName(pVar), ((Double) pObject).doubleValue());
			return "xs:double";
		}
		if (pObject instanceof Calendar) {
			pExpression.bindTimestamp(new QName(pVar), new Timestamp(((Calendar) pObject).getTimeInMillis()));
			return "xs:dateTime";
		}
		if (pObject instanceof Time) {
			pExpression.bindTime(new QName(pVar), new java.sql.Time(((Time) pObject).getCalendar().getTimeInMillis()));
			return "xs:time";
		}
		throw new JAXRException("Unsupported parameter type: " + pObject.getClass().getName());
	}
	
	protected String bindParams(XQExpression pExpression) throws JAXRException, XQException {
	    if (parametersByOffset.isEmpty()) {
	        return getXQuery();
	    }
	    final StringBuilder prolog = new StringBuilder();
		final StringBuilder query = new StringBuilder();
		final String xq = getXQuery();
		int prevOffset = 0;
		for (Parameter param : parametersByOffset) {
			final Object o;
			final String var;
			if (param instanceof NamedParameter) {
				final String name = ((NamedParameter) param).getName();
				o = namedParameterValues.get(name);
				if (o == null) {
					throw new JAXRException("No value bound to parameter " + name);
				}
				var = "param" + Character.toUpperCase(name.charAt(0)) + name.substring(1);
			} else {
				final int num = ((NumberedParameter) param).getNum();
				if (num >= 0  &&  num < numberedParameterValues.size()) {
					o = numberedParameterValues.get(num);
				} else {
					o = null;
				}
				if (o == null) {
					throw new JAXRException("No value bound to parameter " + num);
				}
				var = "param" + num;
			}
			prolog.append("declare variable $");
			prolog.append(var);
			prolog.append(" as ");
			prolog.append(bindParam(pExpression, var, o));
			prolog.append(" external;\n");
			query.append(xq.subSequence(prevOffset, param.getOffset()));
			query.append('$');
			query.append(var);
			prevOffset = param.getOffset() + param.getLength();
		}
		query.append(xq.subSequence(prevOffset, xq.length()));
		final StringBuilder sb = new StringBuilder();
		sb.append(query.subSequence(0, parameterOffset));
		sb.append(prolog);
		sb.append(query.subSequence(parameterOffset, query.length()));
		return sb.toString();
	}
	
	@Override
    public List<RegistryObject[]> executeArrayQuery() throws JAXRException {
		try {
			final XQResultSequence sequence = runQuery(facade);
			final List<String> result = new ArrayList<String>();
			final int dimension = objectTypes.size();
			final Map<String,RegistryObject> keys = fillResult(result, facade, sequence);
			if (result.size() % dimension != 0) {
				throw new JAXRException("Expected result size to be a multiple of "
						+ dimension + ", got " + result.size());
			}
			final List<RegistryObject[]> ros = new ArrayList<RegistryObject[]>(result.size() / dimension);
			for (int i = 0;  i < result.size();  i += dimension) {
				final RegistryObject[] row = new RegistryObject[dimension];
				for (int j = 0;  j < dimension;  j++) {
					row[j] = getRegistryObject(keys, result.get(i+j));
				}
				ros.add(row);
			}
			return ros;
		} catch (XQException e) {
			throw new JAXRException(e);
		}
	}

	@Override
    public List<RegistryObject> executeQuery() throws JAXRException {
		if (objectTypes.size() != 1) {
			throw new JAXRException("This query returns arrays, use executeArrayQuery()");
		}
		try {
			final XQResultSequence sequence = runQuery(facade);
			final List<String> result = new ArrayList<String>();
			final Map<String,RegistryObject> map = fillResult(result, facade, sequence);
			final List<RegistryObject> ros = new ArrayList<RegistryObject>(result.size());
			for (int i = 0;  i < result.size();  i++) {
				final String key = result.get(i);
				ros.add(getRegistryObject(map, key));
			}
			return ros;
		} catch (XQException e) {
			throw new JAXRException(e);
		}
	}

	private RegistryObject getRegistryObject(
			final Map<String, RegistryObject> pMap, final String pKey)
			throws JAXRException {
		final RegistryObject ro;
		if (pKey == null) {
			ro = null;
		} else {
			ro = pMap.get(pKey);
			if (ro == null) {
				throw new JAXRException("No registry object found for key " + pKey);
			}
		}
		return ro;
	}

	private String getObjectTypeName(String pObjectTypeName) {
	    if ("{http://namespaces.csutils.sf.net/query/1.0.0}allTypes".equals(pObjectTypeName)) {
	        return null;
	    }
	    final String mappedObjectTypeName = BUILTIN_NAME_MAP.get(pObjectTypeName);
        return mappedObjectTypeName == null ? pObjectTypeName : mappedObjectTypeName;
	}

	private Map<String,RegistryObject> fillResult(List<String> pResult, RegistryFacade pFacade, XQResultSequence pSequence) throws JAXRException, XQException {
		final BusinessLifeCycleManager blcm = pFacade.getBusinessLifeCycleManager();
		final BusinessQueryManager bqm = pFacade.getBusinessQueryManager();
		while (pSequence.next()) {
			final String key = pSequence.getString();
			pResult.add(key);
		}
		final int dimension = objectTypes.size();
		final Map<String,RegistryObject> map = new HashMap<String,RegistryObject>();
		for (int i = 0;  i < dimension;  i++) {
			final Set<Key> keys = new HashSet<Key>();
			for (int j = i;  j < pResult.size();  j += dimension) {
				keys.add(blcm.createKey(pResult.get(j)));
			}
			final String objectTypeName = getObjectTypeName(objectTypes.get(i).toString());
			final BulkResponse response;
			if (objectTypeName == null) {
			    response = bqm.getRegistryObjects(keys);
			} else {
			    response = bqm.getRegistryObjects(keys, objectTypeName);
			}
			final Collection<RegistryObject> registryObjects = cast(response.getCollection());
			for (RegistryObject ro : registryObjects) {
				map.put(ro.getKey().getId(), ro);
			}
		}
		return map; 
	}
	
	private XQResultSequence runQuery(RegistryFacade pFacade)
			throws XQException, JAXRException {
		final XQExpression expression = ((CentraSiteConnection) pFacade.getConnection()).getXQConnection().createExpression();
		final String q = bindParams(expression);
		log.debug("runQuery: " + q);
		final XQResultSequence sequence = expression.executeQuery(q);
		return sequence;
	}
}
