package net.sf.csutils.core.query.impl;

import java.util.ArrayList;
import java.util.Calendar;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.xml.registry.JAXRException;

import net.sf.csutils.core.model.QName;
import net.sf.csutils.core.model.ROAttribute;
import net.sf.csutils.core.model.ROMetaModel;
import net.sf.csutils.core.query.CsqlStatement;
import net.sf.csutils.core.query.tree.FunctionCall;
import net.sf.csutils.core.query.tree.Identifier;
import net.sf.csutils.core.query.tree.In;
import net.sf.csutils.core.query.tree.IsNull;
import net.sf.csutils.core.query.tree.Like;
import net.sf.csutils.core.query.tree.MethodCall;
import net.sf.csutils.core.query.tree.Negation;
import net.sf.csutils.core.query.tree.SelectStatement;
import net.sf.csutils.core.query.tree.Statement;
import net.sf.csutils.core.registry.centrasite.CentraSiteRegistryInfo;

import org.apache.ws.commons.util.XsDateTimeFormat;


/**
 * Converts a CSQL query into XQuery.
 */
public class XQueryGenerator extends AbstractQueryGenerator {
    private static final XsDateTimeFormat format = new XsDateTimeFormat();
    private Map<String,String> namespaces = new HashMap<String,String>();

    @Override
    protected void parse(Negation pNegation) throws JAXRException {
		final Object item = pNegation.getItem();
		if (item instanceof IsNull) {
			sb.append("exists(");
			parseExpression(((IsNull) item).getItem());
			sb.append(")");
		} else {
			sb.append("not ");
			parseExpression(pNegation.getItem());
		}
	}

	/**
	 * Adds the given string.
	 * @throws JAXRException The string could not be added.
	 */
	@Override
	protected void parse(String pValue) throws JAXRException {
		sb.append("'");
		sb.append(pValue);
		sb.append("'");
	}

	/**
	 * Checks, whether a value is null.
	 * @param pEmpty The empty value
	 * @throws JAXRException Could not add the check for emptiness.
	 */
	@Override
	protected void parse(IsNull pEmpty) throws JAXRException {
		sb.append("not exists(");
		parseExpression(pEmpty.getItem());
		sb.append(")");
	}

	/**
	 * @throws JAXRException The calendar could not be added. 
	 */
	@Override
	protected void parse(Calendar pCalendar) throws JAXRException {
		sb.append('\'');
		sb.append(format.format(pCalendar));
		sb.append('\'');
	}

	/**
	 * @throws JAXRException The calendar could not be added. 
	 */
    @Override
	protected void parse(Boolean pValue) throws JAXRException {
		if (pValue.booleanValue()) {
			sb.append("true()");
		} else {
			sb.append("false()");
		}
	}

    @Override
    protected void parse(FunctionCall pFunctionCall) throws JAXRException {
        final String functionName = pFunctionCall.getFunctionName().toLowerCase();
        final String xQueryFunctionName;
        final int numArgs;
        if ("upper".equals(functionName)) {
            xQueryFunctionName = "upper-case";
            numArgs = 1;
        } else if ("lower".equals(functionName)) {
            xQueryFunctionName = "lower-case";
            numArgs = 1;
        } else {
            throw new JAXRException("Invalid function name: " + functionName);
        }
        final List<Object> args = pFunctionCall.getArgs();
        if (numArgs != args.size()) {
            throw new JAXRException("Expected exactly " + numArgs
                    + " arguments for function " + functionName
                    + ", got " + args.size());
        }
        sb.append(xQueryFunctionName);
        if (args.isEmpty()) {
            sb.append("()");
        } else {
            for (int i = 0;  i < args.size();  i++) {
                sb.append(i == 0 ? '(' : ',');
                parseExpression(args.get(i));
            }
            sb.append(')');
        }
    }

    @Override
    protected void parse(MethodCall pMethodCall) throws JAXRException {
		final String methodName = pMethodCall.getMethodName();
		final String path = pMethodCall.getItem().getPath();
		final List<Object> args = pMethodCall.getArgs();
		int offset = path.indexOf('.');
		if (offset != -1) {
			throw new JAXRException("Invalid method call: Expected <alias>."
					+ methodName + "(...), got " + path);
		}
		if ("key".equals(methodName)
				||  "status".equals(methodName)
				||  "stability".equals(methodName)
				||  "minorVersion".equals(methodName)
				||  "majorVersion".equals(methodName)
				||  "userVersion".equals(methodName)) {
			if (args.size() > 0) {
				throw new JAXRException("The method " + methodName
						+ " allows no arguments.");
			}
			parseAlias(path, path);
			sb.append("/cs:");
			sb.append(methodName);
		} else if ("name".equals(methodName)  ||  "description".equals(methodName)) {
			if (args.size() > 1) {
				throw new JAXRException("The method " + methodName
						+ " must have 0 or 1 argument.");
			}
			parseAlias(path, path);
			sb.append("/cs:");
			sb.append(methodName);
			sb.append("/cs:localString");
			if (!args.isEmpty()) {
				sb.append("[@xml:lang=");
				parseExpression(args.get(0));
				sb.append("]");
			}
		} else {
			throw new JAXRException("Unknown method name: " + methodName);
		}
	}

	protected ObjectTypeInfo parseAlias(String pPath, String pAlias) throws JAXRException {
		final ObjectTypeInfo oti = getAlias(pPath, pAlias);
		sb.append("$");
		sb.append(oti.getAlias());
		return oti;
	}

    @Override
	protected void parse(Like pLike) throws JAXRException {
	    sb.append("sxsLike.sxsLike(");
	    parseExpression(pLike.getItem1());
	    sb.append(',');
        parseExpression(pLike.getItem2());
        sb.append(')');
	}

    @Override
    protected void parse(In pIn) throws JAXRException {
    	parseExpression(pIn.getLhs());
    	sb.append(" = (");
    	final List<?> rhs = pIn.getRhs();
    	for (int i = 0;  i < rhs.size();  i++) {
    		if (i > 0) {
    			sb.append(", ");
    		}
    		parseExpression(rhs.get(i));
    	}
    	sb.append(")");
    }

    @Override
	protected void parse(Identifier pIdentifier) throws JAXRException {
		final String path = pIdentifier.getPath();
		final int offset = path.indexOf('.');
		if (offset <= 0  ||  offset > path.length()-2  ||  path.indexOf('.', offset+1) != -1) {
			throw new JAXRException("Invalid identifier: Expected <alias>.<attribute>, got "
					+ path);
		}
		final String alias = path.substring(0, offset);
		final String attribute = path.substring(offset+1);
		final ObjectTypeInfo oti = parseAlias(path, alias);
		final ROAttribute roAttribute = oti.getROType().getAttribute(attribute);
		if (roAttribute == null) {
			throw new JAXRException("Invalid identifier " + path
					+ ": The object type " + oti.getROType().getQName()
					+ " has no attribute " + attribute);
		}
		switch (roAttribute.getType()) {
        case string:
        case bool:
        case date:
        case dateTime:
        case emailAddress:
        case integer:
        case ipv4Address:
        case ipv6Address:
        case longInteger:
        case number:
        case time:
        case uri:
				break;
			default:
				throw new JAXRException("Invalid identifier " + path
						+ ": attribute type of " + attribute
						+ " in object type " + oti.getROType().getQName()
						+ " is " + roAttribute.getType() + ", which is unsupported in expressions.");
		}
		sb.append('/');
		sb.append(roAttribute.getName());
	}
	
	
	@Override
    protected CsqlStatement parse(SelectStatement pQuery) throws JAXRException {
	    final List<ObjectTypeInfo> otypes = findObjectTypes(pQuery);
        boolean first = true;
        for (ObjectTypeInfo oti : otypes) {
            final QName qName = oti.getROType().getQueryName();
            final String uri = qName.getNamespaceURI();
            String prefix;
            if ("".equals(uri)) {
                prefix = null;
            } else {
                prefix = namespaces.get(qName.getNamespaceURI());
                if (prefix == null) {
                    prefix = "n" + namespaces.size();
                    namespaces.put(qName.getNamespaceURI(), prefix);
                }
            }
            if (first) {
                sb.append("for ");
                first = false;
            } else {
                sb.append(",");
            }
            sb.append("$");
            sb.append(oti.getAlias());
            if (NS_QUERY.equals(qName.getNamespaceURI())) {
                if ("allTypes".equals(qName.getLocalPart())) {
                    sb.append(" in input()/*");
                }
            } else {
                sb.append(" in input()/");
                if (prefix != null) {
                    sb.append(prefix);
                    sb.append(":");
                }
                sb.append(oti.getROType().getQueryName().getLocalPart());
            }
        }
        int parameterOffset = 0;
        for (Map.Entry<String,String> entry : namespaces.entrySet()) {
            final String uri = entry.getKey();
            if (!NS_QUERY.equals(uri)) {
                final String decl = "declare namespace " + entry.getValue() + "='" + uri + "';\n";
                parameterOffset += decl.length();
                sb.insert(0, decl);
            }
        }
        String sep = "\n where ";
        for (int i = 1;  i < otypes.size();  i++) {
            final ObjectTypeInfo oti = otypes.get(i);
            sb.append(sep);
            sep = " and ";
            sb.append("$");
            sb.append(oti.getSourceType().getAlias());
            sb.append("/cs:associations/cs:association[cs:targetObject=$");
            sb.append(oti.getAlias());
            sb.append("/cs:key");
            final String attributeKey = oti.getRelation().getAttributeKey();
            if (attributeKey != null) {
                sb.append(" and cs:instanceSlots/css:attributeKey='");
                sb.append(attributeKey);
                sb.append('\'');
            }
            sb.append("]");
        }
        if (pQuery.getWhereClause() != null) {
            sb.append(sep);
            sep = " and ";
            parseExpression(pQuery.getWhereClause());
        }
        sb.append("\n return ");
        first = true;
        if (otypes.size() > 1) {
            sb.append('(');
        }
        for (ObjectTypeInfo oti : otypes) {
            if (first) {
                first = false;
            } else {
                sb.append(",");
            }
            sb.append("$");
            sb.append(oti.getAlias());
            sb.append("/cs:key/text()");
        }
        if (otypes.size() > 1) {
            sb.append(')');
        }
        final List<QName> objectTypeNames = new ArrayList<QName>(otypes.size());
        for (ObjectTypeInfo oti : otypes) {
            final QName qName = oti.getROType().getQueryName();
            objectTypeNames.add(qName);
        }
        return new XQueryCsqlStatement(sb.toString(), objectTypeNames, getParameters(), parameterOffset);
    }

	@Override
	public CsqlStatement parse(ROMetaModel pModel, Statement pQuery) throws JAXRException {
		namespaces.put(CentraSiteRegistryInfo.NS_SCHEMA_JAXR, "cs");
		namespaces.put(CentraSiteRegistryInfo.NS_SCHEMA, "css");
		return super.parse(pModel, pQuery);
	}
}
