package net.sf.csutils.core.query.tree;

import java.util.List;

import antlr.RecognitionException;


public abstract class NodeFactory {
	public abstract RecognitionException error(String pMessage);
	
	public Parameter parameter() { return new Parameter(); }
	public Parameter namedParameter(String pName) { return new Parameter(pName); }
	public Parameter numberedParameter(String pNumber) throws RecognitionException {
		final int i;
		try {
			i = Integer.parseInt(pNumber);
			return new Parameter(i);
		} catch (Exception e) {
			throw error("Invalid parameter number: " + pNumber);
		}
	}

	/**
	 * Creates a new identifier.
	 * @param pPath The identifiers path.
	 * @return The new identifier.
	 */
	public Identifier identifier(String pPath) {
		return new Identifier(pPath);
	}

	/**
	 * Concatenates the two identifiers.
	 * @param pId1 The prefix part of the identifiers path.
     * @param pId2 The suffix part of the identifiers path.
     * @return The new identifier.
	 */
	public Identifier identifier(Identifier pId1, Identifier pId2) {
		return new Identifier(pId1.getPath() + "." + pId2.getPath());
	}

	/**
	 * Creates a new {@link ObjectType}.
	 * @param pName The object types name.
	 * @param pAlias The object types alias.
	 * @return The created object type.
	 */
	public ObjectType objectType(String pName, String pAlias) {
		return new ObjectType(pName, pAlias);
	}

	public Or or(Object pItem1, Object pItem2) {
		if (pItem1 instanceof Or) {
			final Or item1 = (Or) pItem1;
			if (pItem2 instanceof Or) {
				final List<Object> items2 = ((Or) pItem2).getItems();
				item1.getItems().addAll(items2);
			} else {
				item1.getItems().add(pItem2);
			}
			return item1;
		}
		if (pItem2 instanceof Or) {
			final Or item2 = (Or) pItem2;
			item2.getItems().add(0, pItem1);
			return item2;
		}
		return new Or(pItem1, pItem2);
	}

	public And and(Object pItem1, Object pItem2) {
		if (pItem1 instanceof And) {
			final And item1 = (And) pItem1;
			if (pItem2 instanceof And) {
				final List<Object> items2 = ((And) pItem2).getItems();
				item1.getItems().addAll(items2);
			} else {
				item1.getItems().add(pItem2);
			}
			return item1;
		}
		if (pItem2 instanceof And) {
			final And item2 = (And) pItem2;
			item2.getItems().add(0, pItem1);
			return item2;
		}
		final And result = new And(pItem1, pItem2);
		return result;
	}

	public Object negate(Object pItem) {
		if (pItem instanceof Negation) {
			return ((Negation) pItem).getItem();
		}
		return new Negation(pItem);
	}


	public Object unaryMinus(Object pItem) {
		if (pItem instanceof UnaryMinus) {
			return ((UnaryMinus) pItem).getItem();
		}
		if (pItem instanceof Number) {
			if (pItem instanceof Integer) {
				return Integer.valueOf(-((Integer) pItem).intValue());
			}
			if (pItem instanceof Long) {
				return Long.valueOf(-((Long) pItem).longValue());
			}
			if (pItem instanceof Float) {
				return Float.valueOf(-((Float) pItem).floatValue());
			}
			if (pItem instanceof Double) {
				return Double.valueOf(-((Double) pItem).doubleValue());
			}
		}
		return new UnaryMinus(pItem);
	}

	/**
	 * Creates a check for equality.
     * @param pItem1 The left hand side of the EQ clause.
     * @param pItem2 The right hand side of the EQ clause.
     * @return The created EQ clause.
	 */
	public Eq eq(Object pItem1, Object pItem2) {
		return new Eq(pItem1, pItem2);
	}

	/**
	 * Creates a check for LIKE.
	 * @param pItem1 The left hand side of the LIKE clause.
     * @param pItem2 The right hand side of the LIKE clause.
     * @return The created LIKE clause.
	 */
	public Like like(Object pItem1, Object pItem2) {
		return new Like(pItem1, pItem2);
	}

	/**
	 * Creates a check for NOT LIKE.
     * @param pItem1 The left hand side of the NOT LIKE clause.
     * @param pItem2 The right hand side of the NOT LIKE clause.
     * @return The created NOT LIKE clause.
	 */
	public Negation notLike(Object pItem1, Object pItem2) {
		return new Negation(like(pItem1, pItem2));
	}

	/**
	 * Creates a check for non-equality.
     * @param pItem1 The left hand side of the NE clause.
     * @param pItem2 The right hand side of the NE clause.
     * @return The created NE clause.
	 */
	public Ne ne(Object pItem1, Object pItem2) {
		return new Ne(pItem1, pItem2);
	}

	/**
	 * Creates a check for "lower than".
     * @param pItem1 The left hand side of the LT clause.
     * @param pItem2 The right hand side of the LT clause.
     * @return The created LT clause.
	 */
	public Lt lt(Object pItem1, Object pItem2) {
		return new Lt(pItem1, pItem2);
	}

	/**
	 * Creates a check for "lower than, or equal to".
     * @param pItem1 The left hand side of the LE clause.
     * @param pItem2 The right hand side of the LE clause.
     * @return The created LE clause.
	 */
	public Le le(Object pItem1, Object pItem2) {
		return new Le(pItem1, pItem2);
	}

	/**
	 * Creates a check for "greater than".
     * @param pItem1 The left hand side of the GT clause.
     * @param pItem2 The right hand side of the GT clause.
     * @return The created GT clause.
	 */
	public Gt gt(Object pItem1, Object pItem2) {
		return new Gt(pItem1, pItem2);
	}

	/**
	 * Creates a check for "greater than, or equal to".
     * @param pItem1 The left hand side of the GE clause.
     * @param pItem2 The right hand side of the GE clause.
     * @return The created GE clause.
	 */
	public Ge ge(Object pItem1, Object pItem2) {
		return new Ge(pItem1, pItem2);
	}

	/**
	 * Creates the concatenation of the given items
     * @param pItem1 The left hand side of the concatentation.
     * @param pItem2 The right hand side of the concatenation.
     * @return The created concatenation.
	 */
	public Concatenation concatenate(Object pItem1, Object pItem2) {
		if (pItem1 instanceof Concatenation) {
			final Concatenation item1 = (Concatenation) pItem1;
			if (pItem2 instanceof Concatenation) {
				final List<Object> items2 = ((Concatenation) pItem2).getItems();
				item1.getItems().addAll(items2);
			} else {
				item1.getItems().add(pItem2);
			}
			return item1;
		}
		if (pItem2 instanceof Concatenation) {
			final Concatenation item2 = (Concatenation) pItem2;
			item2.getItems().add(pItem1);
			return item2;
		}
		return new Concatenation(pItem1, pItem2);
	}

	public AdditiveExpression addExpression(Object pItem1, Object pItem2) {
		if (pItem1 instanceof AdditiveExpression) {
			final AdditiveExpression item1 = (AdditiveExpression) pItem1;
			final List<Object> items1 = item1.getItems();
			items1.add(AdditiveExpression.PLUS);
			if (pItem2 instanceof AdditiveExpression) {
				final List<Object> items2 = ((AdditiveExpression) pItem2).getItems();
				item1.getItems().addAll(items2);
			} else {
				item1.getItems().add(pItem2);
			}
			return item1;
		}
		if (pItem2 instanceof AdditiveExpression) {
			final AdditiveExpression item2 = (AdditiveExpression) pItem2;
			final List<Object> items2 = item2.getItems();
			items2.add(0, pItem1);
			items2.add(1, AdditiveExpression.PLUS);
			return item2;
		}
		return new AdditiveExpression(pItem1, AdditiveExpression.PLUS, pItem2);
	}

	public AdditiveExpression minusExpression(Object pItem1, Object pItem2) {
		if (pItem1 instanceof AdditiveExpression) {
			final AdditiveExpression item1 = (AdditiveExpression) pItem1;
			final List<Object> items1 = item1.getItems();
			items1.add(AdditiveExpression.MINUS);
			if (pItem2 instanceof AdditiveExpression) {
				final List<Object> items2 = ((AdditiveExpression) pItem2).getItems();
				invertAdditive(items2);
				item1.getItems().addAll(items2);
			} else {
				item1.getItems().add(pItem2);
			}
			return item1;
		}
		if (pItem2 instanceof AdditiveExpression) {
			final AdditiveExpression item2 = (AdditiveExpression) pItem2;
			final List<Object> items2 = item2.getItems();
			invertAdditive(items2);
			items2.add(0, pItem1);
			items2.add(1, AdditiveExpression.MINUS);
			return item2;
		}
		return new AdditiveExpression(pItem1, AdditiveExpression.MINUS, pItem2);
	}

	private void invertAdditive(final List<Object> items2) {
		for (int i = 1;  i < items2.size();  i += 2) {
			final Object op = items2.get(i);
			if (AdditiveExpression.PLUS.equals(op)) {
				items2.set(i, AdditiveExpression.MINUS);
			} else {
				items2.set(i, AdditiveExpression.PLUS);
			}
		}
	}

	public MultiplicativeExpression multiplyExpression(Object pItem1, Object pItem2) {
		if (pItem1 instanceof MultiplicativeExpression) {
			final MultiplicativeExpression item1 = (MultiplicativeExpression) pItem1;
			final List<Object> items1 = item1.getItems();
			items1.add(MultiplicativeExpression.MULTIPLY);
			if (pItem2 instanceof MultiplicativeExpression) {
				final List<Object> items2 = ((MultiplicativeExpression) pItem2).getItems();
				item1.getItems().addAll(items2);
			} else {
				item1.getItems().add(pItem2);
			}
			return item1;
		}
		if (pItem2 instanceof MultiplicativeExpression) {
			final MultiplicativeExpression item2 = (MultiplicativeExpression) pItem2;
			final List<Object> items2 = item2.getItems();
			items2.add(0, pItem1);
			items2.add(1, MultiplicativeExpression.MULTIPLY);
			return item2;
		}
		return new MultiplicativeExpression(pItem1, MultiplicativeExpression.MULTIPLY, pItem2);
	}

	public MultiplicativeExpression divideExpression(Object pItem1, Object pItem2) {
		if (pItem1 instanceof MultiplicativeExpression) {
			final MultiplicativeExpression item1 = (MultiplicativeExpression) pItem1;
			final List<Object> items1 = item1.getItems();
			items1.add(MultiplicativeExpression.DIVIDE);
			if (pItem2 instanceof MultiplicativeExpression) {
				final List<Object> items2 = ((MultiplicativeExpression) pItem2).getItems();
				invertMultiplicative(items2);
				item1.getItems().addAll(items2);
			} else {
				item1.getItems().add(pItem2);
			}
			return item1;
		}
		if (pItem2 instanceof MultiplicativeExpression) {
			final MultiplicativeExpression item2 = (MultiplicativeExpression) pItem2;
			final List<Object> items2 = item2.getItems();
			invertMultiplicative(items2);
			items2.add(0, pItem1);
			items2.add(1, MultiplicativeExpression.DIVIDE);
			return item2;
		}
		return new MultiplicativeExpression(pItem1, MultiplicativeExpression.DIVIDE, pItem2);
	}

	private void invertMultiplicative(final List<Object> items2) {
		for (int i = 1;  i < items2.size();  i += 2) {
			final Object op = items2.get(i);
			if (MultiplicativeExpression.MULTIPLY.equals(op)) {
				items2.set(i, MultiplicativeExpression.DIVIDE);
			} else {
				items2.set(i, MultiplicativeExpression.MULTIPLY);
			}
		}
	}

	public Integer integerConstant(String pNumber) throws RecognitionException {
		try {
			return Integer.valueOf(pNumber);
		} catch (Exception e) {
			throw error("Invalid integer constant: " + pNumber);
		}
	}

	public Float floatConstant(String pNumber) throws RecognitionException {
		try {
			return Float.valueOf(pNumber);
		} catch (Exception e) {
			throw error("Invalid float constant: " + pNumber);
		}
	}

	public Long longConstant(String pNumber) throws RecognitionException {
		try {
			return Long.valueOf(pNumber);
		} catch (Exception e) {
			throw error("Invalid long constant: " + pNumber);
		}
	}

	public Double doubleConstant(String pNumber) throws RecognitionException {
		try {
			return Double.valueOf(pNumber);
		} catch (Exception e) {
			throw error("Invalid double constant: " + pNumber);
		}
	}

	public String stringConstant(String pValue) throws RecognitionException {
		if (pValue != null  &&  pValue.length() >= 2) {
			final char c1 = pValue.charAt(0);
			if (c1 == '\'') {
				final char c2 = pValue.charAt(pValue.length()-1);
				if (c1 == c2) {
					return pValue.substring(1, pValue.length()-1);
				}
			}
		}
		throw error("Invalid string constant: " + pValue);
	}

	public Boolean trueConstant() {
		return Boolean.TRUE;
	}

	public Boolean falseConstant() {
		return Boolean.FALSE;
	}

	/**
	 * Adds the given object type (aka, a JOIN clause) to the select
	 * statements list of object types.
	 * @param pStatement The select statement to modify
	 * @param pObjectType The JOIN clause to add.
	 * @throws RecognitionException The JOIN clauses alias isn't unique.
	 */
	public void objectType(SelectStatement pStatement, ObjectType pObjectType) throws RecognitionException {
		final String alias = pObjectType.getAlias();
		for (ObjectType ot : pStatement.getObjectTypes()) {
			if (alias.equals(ot.getAlias())) {
				throw error("Non-unique alias: " + alias);
			}
		}
		pStatement.add(pObjectType);
	}

	public IsNull isNull(Object pObject) {
		return new IsNull(pObject);
	}

	public IsEmpty isEmpty(Object pObject) {
		return new IsEmpty(pObject);
	}

	public Object methodCall(Identifier pId, List<Object> pArgs) {
		final String path = pId.getPath();
		int offset = path.lastIndexOf('.');
		if (offset <= 0) {
		    return new FunctionCall(path, pArgs);
		}
		return new MethodCall(new Identifier(path.substring(0, offset)), path.substring(offset+1), pArgs);
	}

	/**
	 * Creates a {@link SelectStatement} with the given set of
	 * namespace declarations.
	 * @param pNamespaceDeclarations The set of namespace declarations to use.
	 * @return The created select statement.
	 */
	public SelectStatement selectStatement(NamespaceDeclarationSet pNamespaceDeclarations) {
		final SelectStatement st = new SelectStatement();
		st.setNamespaces(pNamespaceDeclarations);
		return st;
	}

	/**
	 * Returns a set of namespace declarations, which include a mapping for
	 * the given prefix and URI.
	 * @param pSet The namespace declaration set to modify, if any, or null, if a new set
	 *   should be created.
	 * @param pPrefix The new namespace declarations prefix.
     * @param pUri The new namespace declarations URI.
     * @return The created or modified set of namespace declarations.
	 */
	public NamespaceDeclarationSet namespaceDeclaration(NamespaceDeclarationSet pSet,
			String pPrefix, String pUri) {
		final NamespaceDeclarationSet set = pSet == null ? new NamespaceDeclarationSet() : pSet;
		set.declarePrefix(pPrefix, pUri);
		return set;
	}

	public Identifier allTypes() {
	    return new Identifier("*");
	}

	public In in(Object pLhs, List<?> pRhs, boolean pNot) {
		return new In(pLhs, pRhs, pNot);
	}
}
