package net.sf.csutils.core.registry;

import java.util.Collection;

import javax.xml.registry.JAXRException;
import javax.xml.registry.infomodel.ClassificationScheme;
import javax.xml.registry.infomodel.Concept;
import javax.xml.registry.infomodel.RegistryObject;


/**
 * <p>A utility class for working with concepts. This class
 * provides only static methods and cannot be instantiated.</p>
 * <p>This class supports an extended path notation. A complete
 * path includes the classification schemes, or taxonomies name.
 * For example: <code>//taxonomyName/conceptA/conceptB</code>.</p>
 * <p>In the example, the double slashes indicate that the first
 * component in the path is in fact a taxonomies, or classification
 * schemes name.</p>
 */
public class Concepts {
	/**
	 * Returns the classification scheme, or concept, with the
	 * given path.
	 * @param pFacade The registry facade to use.
	 * @param pPath The path to evaluate.
	 * @return The classification scheme, or concept, with the
     * given path
     * @throws JAXRException The operation failed.
	 */
	public static RegistryObject findObjectByPath(RegistryFacade pFacade, String pPath) throws JAXRException {
		if (pPath.startsWith("//")) {
			return findObjectByPath(pFacade, pPath, 2);
		} else if (pPath.startsWith("/")) {
			/* Upwards compatibility: We treat the path as if
			 * the user had double slashes as a prefix.
			 */
			return findObjectByPath(pFacade, pPath, 1);
		} else {
			throw new IllegalArgumentException("Invalid absolute path: " + pPath);
		}
	}

	private static RegistryObject findObjectByPath(RegistryFacade pFacade,
			String pPath, final int pIndex) throws JAXRException {
		final int offset = pPath.indexOf('/', pIndex);
		if (offset == -1) {
			return pFacade.findClassificationScheme(pPath.substring(pIndex));
		}
		final String relativePath = pPath.substring(offset+1);
		final ClassificationScheme cs = pFacade.findClassificationScheme(pPath.substring(pIndex, offset));
		return findConceptByPath(pFacade, cs, relativePath);
	}

	/**
	 * Returns the concept with the given path, relative to the
	 * given registry object.
	 * @param pFacade The registry facade to use
	 * @param pObject The object, relative to which the path should be evaluated.
	 * @param pPath The path to evaluate.
	 * @return The concept with the given path, relative to the
     * given registry object.
     * @throws JAXRException The operation failed.
	 */
	public static Concept findConceptByPath(RegistryFacade pFacade, RegistryObject pObject, String pPath) throws JAXRException {
		int offset = pPath.indexOf('/');
		if (offset == -1) {
			return pFacade.findDescendant(pObject, pPath);
		}
		final String segment = pPath.substring(0, offset);
		final String path = pPath.substring(offset+1);
		final Concept concept = pFacade.findDescendant(pObject, segment);
		if (concept == null) {
			return null;
		}
		return findConceptByPath(pFacade, concept, path);
	}

	/**
	 * Returns the given objects children concepts.
	 * @param pObject The object to query for children concepts.
	 * @return The objects children concepts.
	 * @throws IllegalArgumentException The object is null, or another
	 *   object than a {@link ClassificationScheme}, or {@link Concept}.
	 *   In other words, it doesn't have any children concepts.
	 * @throws JAXRException The operation failed.
	 */
	public static Collection<Concept> getChildrenConcepts(RegistryObject pObject)
			throws JAXRException {
		if (pObject == null) {
			throw new IllegalArgumentException("The parent object must not be null.");
		}
		final Collection<?> concepts;
		if (pObject instanceof Concept) {
			concepts = ((Concept) pObject).getChildrenConcepts();
		} else if (pObject instanceof ClassificationScheme) {
			concepts = ((ClassificationScheme) pObject).getChildrenConcepts();
		} else {
			throw new IllegalArgumentException("The parent obejct must be a ClassificationScheme "
					+ ", or Concept, but is a " + pObject.getClass().getName());
		}
		return asConceptCollection(concepts);
	}

	/**
	 * Casts the given collection to a collection of concepts.
	 * @param pCollection The concept collection to cast.
	 * @return The casted collection of concepts.
	 */
	@SuppressWarnings("unchecked")
	public static Collection<Concept> asConceptCollection(Collection<?> pCollection) {
		return (Collection<Concept>) pCollection;
	}
}
