package net.sf.csutils.core.registry;

import java.io.Serializable;
import java.net.PasswordAuthentication;
import java.net.URL;
import java.util.HashSet;
import java.util.Properties;

import javax.xml.registry.Connection;
import javax.xml.registry.ConnectionFactory;
import javax.xml.registry.JAXRException;


/**
 * Default implementation of {@link ConnectionProvider}, suitable for
 * standalone programs, which is configured with a registry URL, a
 * user name, and a password.
 */
public class DefaultConnectionProvider implements ConnectionProvider, Serializable {
	private static final long serialVersionUID = 7904241784005341100L;
	private final URL url;
	private final String user, password;
	private final boolean isManagingTypes;

	/**
	 * Creates a new instance with the given registry URL, user
	 * name, and password. The returned connection is not suitable
	 * for using the CentraSite type management API.
	 * @param pURL The URL of the CentraSite database
	 * @param pUser The user name to use for logging in to the CentraSite database.
     * @param pPassword The password to use for logging in to the CentraSite database.
	 */
	public DefaultConnectionProvider(URL pURL, String pUser, String pPassword) {
		this(pURL, pUser, pPassword, false);
	}

	/**
	 * Creates a new instance with the given registry URL, user
	 * name, and password.
     * @param pURL The URL of the CentraSite database
     * @param pUser The user name to use for logging in to the CentraSite database.
     * @param pPassword The password to use for logging in to the CentraSite database.
	 * @param pManagingTypes Whether the returned connection must be
	 *   suitable for using the CentraSite type management API.
	 */
	public DefaultConnectionProvider(URL pURL, String pUser, String pPassword,
			boolean pManagingTypes) {
		url = pURL;
		user = pUser;
		password = pPassword;
		isManagingTypes = pManagingTypes;
	}

	@Override
	public Connection getConnection() throws JAXRException {
        System.setProperty("javax.xml.registry.ConnectionFactoryClass",
        	"com.centrasite.jaxr.ConnectionFactoryImpl");
        Properties p = new Properties();
        p.setProperty("javax.xml.registry.queryManagerURL", url.toExternalForm());
        if (isManagingTypes) {
        	p.put("com.centrasite.jaxr.BrowserBehaviour", "yes");
    	}
        HashSet<PasswordAuthentication> credentials = new HashSet<PasswordAuthentication>();
        credentials.add(new PasswordAuthentication(user, password.toCharArray()));

        final ConnectionFactory connFactory = ConnectionFactory.newInstance();
        connFactory.setProperties(p);
        final Connection connection = connFactory.createConnection();
        connection.setCredentials(credentials);
        connection.setSynchronous(true);
        return connection;
	}
}
