package net.sf.csutils.core.registry;

import java.util.Collection;

import javax.xml.registry.BusinessLifeCycleManager;
import javax.xml.registry.BusinessQueryManager;
import javax.xml.registry.Connection;
import javax.xml.registry.JAXRException;
import javax.xml.registry.infomodel.Association;
import javax.xml.registry.infomodel.ClassificationScheme;
import javax.xml.registry.infomodel.Concept;
import javax.xml.registry.infomodel.RegistryObject;

import net.sf.csutils.core.model.QName;


/**
 * <p>A registry facade acts as a wrapper for the CentraSite
 * registry. The purpose of the facade is caching as well
 * as hiding some of the more complex details of the
 * registry.</p>
 * <p>Depending on your use case, there are different flavours
 * of the facade and you must choose one suitable for you:
 * <ol>
 *   <li>The {@link SimpleRegistryFacade} is a per-connection
 *     object. It is closely tied to a single registry
 *     connection. Its lifetime begins with opening the
 *     connection and ends when the connection is closed.</li>
 * </ol>
 * Other implementations might be possible in the future.</p>
 */
public interface RegistryFacade {
    /**
     * Returns the {@link RegistryInfo registry info object}.
     * @return The {@link RegistryInfo registry info object}.
     */
    RegistryInfo getRegistryInfo();
    
    /**
     * Returns the {@link BusinessQueryManager}.
     * @return The {@link BusinessQueryManager} to use.
     * @throws JAXRException The operation failed.
     */
    BusinessQueryManager getBusinessQueryManager() throws JAXRException;

    /**
     * Returns the {@link BusinessLifeCycleManager}.
     * @return The {@link BusinessLifeCycleManager}.
     * @throws JAXRException The operation failed.
     */
    BusinessLifeCycleManager getBusinessLifeCycleManager() throws JAXRException;

    /**
     * Returns the {@link Connection registry connection}.
     * @return The registry connection to use.
     * @throws JAXRException The operation failed.
     */
    Connection getConnection() throws JAXRException;

    /**
     * Returns the classification scheme, or taxonomy, with the given name.
     * @param pName The classification schemes name.
     * @return The classification scheme, or taxonomy, with the given name.
     * @throws JAXRException The operation failed.
     */
    ClassificationScheme findClassificationScheme(String pName) throws JAXRException;

    /**
     * Returns the classification scheme, or concept with the
     * given path. The notation of a path follows the scheme
     * as outlined in {@link Concepts}.
     * @param pPath The path to evaluate.
     * @return The classification scheme, or concept with the given path.
     * @throws JAXRException The operation failed.
     */
    RegistryObject findObjectByPath(String pPath) throws JAXRException;

    /**
     * Returns the concept with the given path, relative to the given
     * object.
     * @param pParent The registry object, relative to which the path should be evaluated.
     * @param pPath The path to evaluate.
     * @return The concept with the given path, relative to the given.
     * @throws IllegalArgumentException The parent object or path are null.
     * @throws JAXRException The operation failed.
     */
    Concept findDescendant(RegistryObject pParent, String pPath) throws JAXRException;

    /**
     * Returns the {@link Concept association type} with the given name.
     * @param pName The association types name.
     * @return The {@link Concept association type} with the given name.
     * @throws JAXRException The operation failed.
     */
    Concept findAssociationType(String pName) throws JAXRException;

    /**
     * Returns the {@link Concept object type} with the given name.
     * @param pName The object types name.
     * @return The {@link Concept object type} with the given name.
     * @throws JAXRException The operation failed.
     */
    Concept findObjectType(QName pName) throws JAXRException;

    /**
     * Returns the {@link Concept object type} with the given name.
     * @param pName The object types name.
     * @return The {@link Concept object type} with the given name.
     * @throws JAXRException The operation failed.
     */
    Concept findObjectType(String pName) throws JAXRException;

    /**
     * Returns, whether the given concept is an association type concept.
     * @param pConcept The concept to evaluate
     * @return True, if the concept is an association type concept.
     * @throws JAXRException The operation failed.
     */
    boolean isAssociationTypeConcept(Concept pConcept) throws JAXRException;

    /**
     * Returns, whether the given concept is an object type concept.
     * @param pConcept The concept to evaluate
     * @return True, if the concept is an object type concept.
     * @throws JAXRException The operation failed.
     */
    boolean isObjectTypeConcept(Concept pConcept) throws JAXRException;

    /**
     * Returns, whether the given classification scheme is the association type classification scheme.
     * @param pClassificationScheme The classification scheme to evaluate.
     * @return True, if the classification scheme is the taxonomy of association types.
     * @throws JAXRException The operation failed.
     */
    boolean isAssociationTypeClassificationScheme(ClassificationScheme pClassificationScheme) throws JAXRException;

    /**
     * Returns, whether the given classification scheme is the object type classification scheme.
     * @param pClassificationScheme The classification scheme to evaluate.
     * @return True, if the classification scheme is the taxonomy of object types.
     * @throws JAXRException The operation failed.
     */
    boolean isObjectTypeClassificationScheme(ClassificationScheme pClassificationScheme) throws JAXRException;

    /**
     * Returns the classification scheme of association types.
     * @return The classification scheme of association types.
     * @throws JAXRException The operation failed.
     */
    ClassificationScheme getAssociationTypeClassificationScheme() throws JAXRException;

    /**
     * Returns the classification scheme of object types.
     * @return The classification scheme of object types.
     * @throws JAXRException The operation failed.
     */
    ClassificationScheme getObjectTypeClassificationScheme() throws JAXRException;

    /**
     * Returns all associations with the given object as the associations source.
     */
    Collection<Association> getReferencingAssociations(RegistryObject pRo) throws JAXRException;
}
