package net.sf.csutils.core.registry;

import java.util.Collection;
import java.util.Collections;

import javax.xml.registry.BusinessLifeCycleManager;
import javax.xml.registry.BusinessQueryManager;
import javax.xml.registry.Connection;
import javax.xml.registry.FindQualifier;
import javax.xml.registry.JAXRException;
import javax.xml.registry.RegistryService;
import javax.xml.registry.infomodel.ClassificationScheme;
import javax.xml.registry.infomodel.Concept;
import javax.xml.registry.infomodel.RegistryObject;

import net.sf.csutils.core.registry.impl.AbstractRegistryFacade;


/**
 * This is the most simple kind of {@link RegistryFacade}: It
 * doesn't hold any internal cache. All methods are mapped
 * directly to calls of the {@link BusinessQueryManager}.
 */
public class SimpleRegistryFacade extends AbstractRegistryFacade {
	private static final Collection<String> exactNameMatch = Collections.singleton(FindQualifier.EXACT_NAME_MATCH);

	private final Connection connection;
	private final BusinessQueryManager bqm;
	private final BusinessLifeCycleManager blcm;

	/**
	 * Creates a new instance with the given registry connection
	 * and the default {@link RegistryInfo}.
	 * @param pConnection The connection to use.
	 * @throws JAXRException The operation failed.
	 */
	public SimpleRegistryFacade(Connection pConnection) throws JAXRException {
		this(RegistryInfo.getInstance(), pConnection);
	}

	/**
	 * Creates a new instance with the given registry connection
	 * and the given {@link RegistryInfo}.
	 * @param pRegistryInfo The registry information to use.
     * @param pConnection The connection to use.
     * @throws JAXRException The operation failed.
	 */
	public SimpleRegistryFacade(RegistryInfo pRegistryInfo, Connection pConnection) throws JAXRException {
		super(pRegistryInfo);
		connection = pConnection;
		final RegistryService rs = pConnection.getRegistryService();
		bqm = rs.getBusinessQueryManager();
		blcm = rs.getBusinessLifeCycleManager();
	}

	/**
	 * Returns the {@link BusinessQueryManager}.
	 */
	@Override
	public BusinessQueryManager getBusinessQueryManager() {
		return bqm;
	}

	/**
	 * Returns the {@link BusinessLifeCycleManager}.
	 */
	@Override
	public BusinessLifeCycleManager getBusinessLifeCycleManager() {
		return blcm;
	}

	/**
	 * Returns the {@link Connection registry connection}.
	 */
	@Override
	public Connection getConnection() {
		return connection;
	}

	@Override
	public ClassificationScheme findClassificationScheme(String pName)
			throws JAXRException {
		return getBusinessQueryManager().findClassificationSchemeByName(exactNameMatch, pName);
	}

	@Override
	public RegistryObject findObjectByPath(String pPath) throws JAXRException {
		return Concepts.findObjectByPath(this, pPath);
	}

	@Override
	public Concept findDescendant(RegistryObject pParent, String pPath) throws JAXRException {
		if (pParent == null) {
			throw new IllegalArgumentException("The parent object must not be null.");
		}
		if (pPath == null) {
			throw new IllegalArgumentException("The concept path must not be null.");
		}
		for (Concept concept : Concepts.getChildrenConcepts(pParent)) {
			if (pPath.equals(concept.getValue())) {
				return concept;
			}
		}
		return null;
	}

	@Override
	public Concept findAssociationType(String pName) throws JAXRException {
		return findDescendant(getAssociationTypeClassificationScheme(), pName);
	}

	@Override
	public Concept findObjectType(String pName) throws JAXRException {
		return findDescendant(getObjectTypeClassificationScheme(), pName);
	}
}
