package net.sf.csutils.core.registry.centrasite;

import static net.sf.csutils.core.utils.Generics.cast;

import java.net.PasswordAuthentication;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.xml.registry.Connection;
import javax.xml.registry.JAXRException;
import javax.xml.registry.LifeCycleManager;
import javax.xml.registry.infomodel.ClassificationScheme;
import javax.xml.registry.infomodel.Concept;

import net.sf.csutils.core.model.QName;
import net.sf.csutils.core.model.ROAttribute;
import net.sf.csutils.core.model.ROClassification;
import net.sf.csutils.core.model.ROFile;
import net.sf.csutils.core.model.RORelation;
import net.sf.csutils.core.model.ROSlot;
import net.sf.csutils.core.model.ROType;
import net.sf.csutils.core.model.impl.StaticROAttribute;
import net.sf.csutils.core.model.impl.StaticROClassification;
import net.sf.csutils.core.model.impl.StaticROFile;
import net.sf.csutils.core.model.impl.StaticRORelation;
import net.sf.csutils.core.model.impl.StaticROSlot;
import net.sf.csutils.core.model.impl.StaticROType;
import net.sf.csutils.core.registry.RegistryFacade;
import net.sf.csutils.core.registry.impl.MetaModelAccessor;
import net.sf.csutils.core.utils.RegistryObjects;

import org.apache.log4j.Logger;

import com.centrasite.jaxr.CentraSiteConnection;
import com.centrasite.jaxr.CentraSiteLifeCycleManager;
import com.centrasite.jaxr.CentraSiteQueryManager;
import com.centrasite.jaxr.assettype.AssetTypeManager;
import com.centrasite.jaxr.type.CentraSiteAttribute;
import com.centrasite.jaxr.type.CentraSiteClassificationAttribute;
import com.centrasite.jaxr.type.CentraSiteFileAttribute;
import com.centrasite.jaxr.type.CentraSiteRelationShipAttribute;
import com.centrasite.jaxr.type.CentraSiteSlotDescription;
import com.centrasite.jaxr.type.CentraSiteSlotType;
import com.centrasite.jaxr.type.CentraSiteTypeDescription;
import com.centrasite.util.RepositoryConnector;
import com.centrasite.util.RepositoryConnectorFactory;
import com.softwareag.cst.util.base.FileUtil;
import com.softwareag.cst.util.security.Credentials;


/**
 * CentraSite specific implementation of {@link MetaModelAccessor}.
 */
public class CentraSiteMetaModelAccessor extends MetaModelAccessor {
    private static final Logger log = Logger.getLogger(CentraSiteMetaModelAccessor.class);
    private static final Set<QName> BUILTIN_TYPES = new HashSet<QName>(Arrays.asList(new QName[]{
        new QName(LifeCycleManager.ASSOCIATION),
        new QName(LifeCycleManager.SERVICE_BINDING),
        new QName(LifeCycleManager.EXTERNAL_IDENTIFIER),
        new QName(LifeCycleManager.CLASSIFICATION),
        new QName(LifeCycleManager.SPECIFICATION_LINK),
        new QName(LifeCycleManager.AUDITABLE_EVENT)
    }));
    public static final Map<QName,QName> BUILTIN_NAME_MAP = new HashMap<QName,QName>();
    static {
        BUILTIN_NAME_MAP.put(new QName("ClassificationScheme"), new QName(CentraSiteRegistryInfo.NS_SCHEMA_JAXR, "classificationScheme"));
        BUILTIN_NAME_MAP.put(new QName("CentraSiteFilter"), new QName(CentraSiteRegistryInfo.NS_SCHEMA_JAXR, "centraSiteFilter"));
        BUILTIN_NAME_MAP.put(new QName("CentraSiteOperation"), new QName(CentraSiteRegistryInfo.NS_SCHEMA_JAXR, "centraSiteOperation"));
        BUILTIN_NAME_MAP.put(new QName("Concept"), new QName(CentraSiteRegistryInfo.NS_SCHEMA_JAXR, "concept"));
        BUILTIN_NAME_MAP.put(new QName("ExternalLink"), new QName(CentraSiteRegistryInfo.NS_SCHEMA_JAXR, "externalLink"));
        BUILTIN_NAME_MAP.put(new QName("Organization"), new QName(CentraSiteRegistryInfo.NS_SCHEMA_JAXR, "organization"));
        BUILTIN_NAME_MAP.put(new QName("Package"), new QName(CentraSiteRegistryInfo.NS_SCHEMA_JAXR, "registryPackage"));
        BUILTIN_NAME_MAP.put(new QName("RegistryEntry"), new QName(CentraSiteRegistryInfo.NS_SCHEMA_JAXR, "registryEntry"));
        BUILTIN_NAME_MAP.put(new QName("Service"), new QName(CentraSiteRegistryInfo.NS_SCHEMA_JAXR, "service"));
        BUILTIN_NAME_MAP.put(new QName("User"), new QName(CentraSiteRegistryInfo.NS_SCHEMA_JAXR, "user"));
    }
    private static final CentraSiteMetaModelAccessor theInstance = new CentraSiteMetaModelAccessor();

    /**
     * Private constructor, to ensure singleton pattern.
     */
    private CentraSiteMetaModelAccessor() {
        // Does nothing
    }
    
    /**
     * Returns the singleton instance.
     * @return The singleton instance.
     */
    public static CentraSiteMetaModelAccessor getInstance() {
        return theInstance;
    }

    @Override
    public MetaModel newMetaModel(RegistryFacade pFacade) {
        return new MetaModel(pFacade);
    }

    @Override
    public ROType getROType(MetaModel pModel, QName pQName) throws JAXRException {
        synchronized (pModel) {
            final Map<QName, ROType> types = pModel.getTypes();
            final ROType type = types.get(pQName);
            if (type != null) {
                return type;
            }
            final CentraSiteQueryManager csqm = (CentraSiteQueryManager) pModel.getFacade().getBusinessQueryManager();
            final CentraSiteTypeDescription cstd = csqm.getTypeDescription(pQName.toString());
            if (cstd == null) {
                if (BUILTIN_TYPES.contains(pQName)) {
                    final StaticROType roType = new StaticROType();
                    roType.setQName(pQName);
                    roType.setQueryName(BUILTIN_NAME_MAP.get(pQName));
                    roType.setAttributes(new HashMap<String,ROAttribute>());
                    types.put(pQName, roType);
                    return roType;
                }
                log.error("getROType: Unknown registry object type " + pQName);
                return null;
            }
            final Map<String,ROAttribute> attributes = new HashMap<String,ROAttribute>();
            final StaticROType roType = new StaticROType();
            roType.setQName(pQName);
            roType.setQueryName(BUILTIN_NAME_MAP.get(pQName));
            roType.setAttributes(attributes);
            /* Intentional to put the type into the map first and then add the slots.
             * If we didn't, we might have a problem with circular references.
             */
            types.put(pQName, roType);
            addSlots(cstd, attributes);
            addRelations(pModel, cstd, attributes);
            addClassifications(cstd, attributes);
            addFiles(cstd, attributes);
            return roType;
        }
    }

    private void map(CentraSiteAttribute pFrom, StaticROAttribute pTo) {
        pTo.setName(pFrom.getName());
        pTo.setMinOccurs(Integer.parseInt(pFrom.getMinOccurs()));
        final String maxOccurs = pFrom.getMaxOccurs();
        pTo.setMaxOccurs("unbounded".equals(maxOccurs) ? -1 : Integer.parseInt(maxOccurs));
    }
    
    private void addSlots(final CentraSiteTypeDescription pTypeDescription, Map<String,ROAttribute> pAttributes) {
        final Collection<String> slotNames = cast(pTypeDescription.getSlotNames());
        for (String slotName : slotNames) {
            final CentraSiteSlotDescription cssd = pTypeDescription.getSlotDescription(slotName);
            if (cssd == null) {
                log.error("getROType: Unknown slot name " + slotName);
                continue;
            }
            final ROAttribute.Type roType;
            final String dataType = cssd.getDataType();
            if (CentraSiteSlotType.XS_STRING.equals(dataType)) {
                roType = ROAttribute.Type.string;
            } else if (CentraSiteSlotType.XS_BOOLEAN.equals(dataType)) {
                roType = ROAttribute.Type.bool;
            } else if (CentraSiteSlotType.XS_DATETIME.equals(dataType)) {
                roType = ROAttribute.Type.dateTime;
            } else if (CentraSiteSlotType.XS_DATE.equals(dataType)) {
                roType = ROAttribute.Type.date;
            } else if (CentraSiteSlotType.XS_TIME.equals(dataType)) {
                roType = ROAttribute.Type.time;
            } else if (CentraSiteSlotType.XS_DOUBLE.equals(dataType)) {
                roType = ROAttribute.Type.number;
            } else if (CentraSiteSlotType.EMAILADDRESS.equals(dataType)) {
                roType = ROAttribute.Type.emailAddress;
            } else if (CentraSiteSlotType.IPV4.equals(dataType)) {
                roType = ROAttribute.Type.ipv4Address;
            } else if (CentraSiteSlotType.IPV6.equals(dataType)) {
                roType = ROAttribute.Type.ipv6Address;
            } else if (CentraSiteSlotType.XS_ANYURI.equals(dataType)) {
                roType = ROAttribute.Type.uri;
            } else if (CentraSiteSlotType.XS_INTEGER.equals(dataType)) {
                roType = ROAttribute.Type.longInteger;
            } else if ("xs:int".equals(dataType)) {
                roType = ROAttribute.Type.integer;
            } else if ("xs:long".equals(dataType)) {
                roType = ROAttribute.Type.longInteger;
            } else if (CentraSiteSlotType.XS_DURATION.equals(dataType)) {
                roType = ROAttribute.Type.duration;
            } else {
                log.error("getROType: Unknown data type " + cssd.getDataType()
                        + " for slot name " + slotName);
                continue;
            }
            final StaticROSlot attr = new StaticROSlot();
            map(cssd, attr);
            attr.setType(roType);
            pAttributes.put(cssd.getLocalName(), attr);
        }
    }

    private void addFiles(final CentraSiteTypeDescription pTypeDescription, Map<String,ROAttribute> pAttrs) throws JAXRException {
        for (CentraSiteFileAttribute attr : pTypeDescription.getFileAttributes()) {
            final StaticROFile file = new StaticROFile();
            map(attr, file);
            file.setAttributeKey(attr.getAttributeKey());
            pAttrs.put(file.getName(), file);
        }
    }
    
    private void addClassifications(final CentraSiteTypeDescription pTypeDescription, Map<String,ROAttribute> pAttributes) throws JAXRException {
        for (CentraSiteClassificationAttribute attr : pTypeDescription.getClassificationAttributes()) {
            final StaticROClassification classification = new StaticROClassification();
            map(attr, classification);
            classification.setAttributeKey(attr.getAttributeKey());
            classification.setTaxonomy(RegistryObjects.getName(attr.getTaxonomy()));
            pAttributes.put(attr.getName(), classification);
        }
    }

    private void addRelations(MetaModel pModel, final CentraSiteTypeDescription pTypeDescription, Map<String,ROAttribute> pAttributes) throws JAXRException {
        for (CentraSiteRelationShipAttribute attr : pTypeDescription.getRelationShipAttributes()) {
            final StaticRORelation relation = new StaticRORelation();
            map(attr, relation);
            relation.setAssociationType(attr.getAssociationType().getValue());
            relation.setAttributeKey(attr.getAttributeKey());
            pAttributes.put(attr.getName(), relation);
            final List<ROType> targetTypes = new ArrayList<ROType>();
            for (Concept targetType : attr.getTargetTypes()) {
                final QName qName = QName.valueOf(targetType.getValue());
                final ROType targetROType = getROType(pModel, qName);
                if (targetROType != null) {
                    targetTypes.add(targetROType);
                }
            }
            relation.setTargetTypes(targetTypes);
        }
    }

    private RepositoryConnector getRepositoryConnector(Connection pConn) throws JAXRException {
        final String queryManagerUrl = ((CentraSiteConnection)pConn).getQueryManagerURL();
        final String repUrl = FileUtil.combinePaths(queryManagerUrl, "/ino:dav/ino:dav");
        Credentials credentials = null;
        final Set<?> credSet = pConn.getCredentials();
        final Iterator<?> iterator = credSet.iterator();
        while (iterator.hasNext()) {
            Object obj = iterator.next();
            if (obj instanceof PasswordAuthentication) {
                final PasswordAuthentication passwordAuthentication = (PasswordAuthentication)obj;
                final String userName = passwordAuthentication.getUserName();
                final String password = new String(passwordAuthentication.getPassword());
                credentials = new Credentials(userName, password);
                break;
            }
        }
        if (credentials == null) {
            throw new IllegalStateException("No credentials found.");
        }
        try {
            return RepositoryConnectorFactory.newRepositoryConnector(repUrl, credentials);
        } catch (JAXRException e) {
            throw e;
        } catch (RuntimeException e) {
            throw e;
        } catch (Error e) {
            throw e;
        } catch (Exception e) {
            throw new JAXRException(e);
        }
    }

    @Override
    protected void remove(MetaModel pModel, ROType pType) throws JAXRException {
        final QName qName = pType.getQName();
        final CentraSiteLifeCycleManager cslm = (CentraSiteLifeCycleManager) pModel.getFacade().getBusinessLifeCycleManager();
        final CentraSiteQueryManager csqm = (CentraSiteQueryManager) pModel.getFacade().getBusinessQueryManager();
        final Collection<?> instances =
            csqm.findObjects(null, Collections.singleton(qName.toString()),
                null, null, null, null, null).getCollection();
        cslm.deleteObjects(instances);
        final CentraSiteTypeDescription cstd = csqm.getTypeDescription(qName.toString());
        try {
            cslm.deleteTypes(Collections.singleton(cstd), getRepositoryConnector(pModel.getFacade().getConnection()));
        } catch (RuntimeException e) {
            throw e;
        } catch (Error e) {
            throw e;
        } catch (JAXRException e) {
            throw e;
        } catch (Exception e) {
            throw new JAXRException(e);
        }
    }

    private com.centrasite.jaxr.assettype.CentraSiteAttribute createFileAttribute(AssetTypeManager pAtm, ROFile pAttr) throws JAXRException {
        return pAtm.createFileAttribute(pAttr.getName(), pAttr.getName());
    }

    private com.centrasite.jaxr.assettype.CentraSiteAttribute createClassificationAttribute(MetaModel pModel, AssetTypeManager pAtm, ROClassification pAttr) throws JAXRException {
        final ClassificationScheme scheme = pModel.getFacade().findClassificationScheme(pAttr.getTaxonomy());
        if (scheme == null) {
            throw new JAXRException("No such taxonomy: " + pAttr);
        }
        return pAtm.createClassificationAttribute(pAttr.getName(), pAttr.getName(), scheme);
    }

    private com.centrasite.jaxr.assettype.CentraSiteAttribute createRelationAttribute(MetaModel pModel, AssetTypeManager pAtm, RORelation pAttr) throws JAXRException {
        final RegistryFacade facade = pModel.getFacade();
        final Concept associationType = facade.findAssociationType(pAttr.getAssociationType());
        if (associationType == null) {
            throw new JAXRException("No such association type: " + pAttr.getAssociationType());
        }
        final List<Concept> targetTypes = new ArrayList<Concept>();
        for (ROType targetType : pAttr.getTargetTypes()) {
            final Concept targetTypeConcept = facade.findObjectType(targetType.getQName());
            if (targetTypeConcept == null) {
                throw new JAXRException("No such target type: " + targetType.getQName());
            }
            targetTypes.add(targetTypeConcept);
        }
        com.centrasite.jaxr.assettype.CentraSiteAttribute attr = pAtm.createRelationShipAttribute(pAttr.getName(),
                pAttr.getName(), associationType, targetTypes);
        return attr;
    }

    private com.centrasite.jaxr.assettype.CentraSiteAttribute createSlotAttribute(AssetTypeManager pAtm, ROSlot pAttr) throws JAXRException {
        final String dataType;
        switch (pAttr.getType()) {
            case bool:
                dataType = CentraSiteSlotType.XS_BOOLEAN;
                break;
            case date:
                dataType = CentraSiteSlotType.XS_DATE;
                break;
            case dateTime:
                dataType = CentraSiteSlotType.XS_DATETIME;
                break;
            case duration:
                dataType = CentraSiteSlotType.XS_DURATION;
                break;
            case emailAddress:
                dataType = CentraSiteSlotType.EMAILADDRESS;
                break;
            case ipv4Address:
                dataType = CentraSiteSlotType.IPV4;
                break;
            case ipv6Address:
                dataType = CentraSiteSlotType.IPV6;
                break;
            case integer:
                dataType = CentraSiteSlotType.XS_INTEGER;
                break;
            case longInteger:
                dataType = CentraSiteSlotType.XS_INTEGER;
                break;
            case number:
                dataType = CentraSiteSlotType.XS_DOUBLE;
                break;
            case string:
                dataType = CentraSiteSlotType.XS_STRING;
                break;
            case time:
                dataType = CentraSiteSlotType.XS_TIME;
                break;
            case uri:
                dataType = CentraSiteSlotType.XS_ANYURI;
                break;
            case classification:
            case relation:
            case file:
            default:
                throw new IllegalStateException("Invalid slot type: " + pAttr.getType());
        }
        return pAtm.createAttribute(pAttr.getName(), pAttr.getName(), dataType);
    }
    
    private void addAttribute(MetaModel pMetaModel, AssetTypeManager pAtm, CentraSiteTypeDescription pCstm, ROAttribute pAttr) throws JAXRException {
        final com.centrasite.jaxr.assettype.CentraSiteAttribute attr;
        switch (pAttr.getType()) {
            case relation:
                attr = createRelationAttribute(pMetaModel, pAtm, (RORelation) pAttr);
                break;
            case classification:
                attr = createClassificationAttribute(pMetaModel, pAtm, (ROClassification) pAttr);
                break;
            case file:
                attr = createFileAttribute(pAtm, (ROFile) pAttr);
                break;
            default:
                attr = createSlotAttribute(pAtm, (ROSlot) pAttr);
                break;
        }
        attr.setMaxOccurs(pAttr.getMaxOccurs() == -1 ? "unbounded" : String.valueOf(pAttr.getMaxOccurs()));
        attr.setMinOccurs(String.valueOf(pAttr.getMinOccurs()));
        pAtm.addAttribute(pCstm, attr);
    }

    @Override
    public ROType add(MetaModel pModel, ROType pType) throws JAXRException {
        final AssetTypeManager atm = new AssetTypeManager(pModel.getFacade().getConnection());
        final QName qName = pType.getQName();
        final String localPart = qName.getLocalPart();
        final CentraSiteTypeDescription cstd = atm.createAssetType(localPart, localPart, localPart, qName.getNamespaceURI());
        for (ROAttribute attr : pType.getAttributes().values()) {
            addAttribute(pModel, atm, cstd, attr);
        }
        try {
            RepositoryConnector repositoryConnector = getRepositoryConnector(pModel.getFacade().getConnection());
            cstd.save(repositoryConnector);
        } catch (RuntimeException e) {
            throw e;
        } catch (Error e) {
            throw e;
        } catch (JAXRException e) {
            throw e;
        } catch (Exception e) {
            throw new JAXRException(e);
        }
        return pType;
    }
}
