package net.sf.csutils.core.registry.impl;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

import javax.xml.registry.BusinessLifeCycleManager;
import javax.xml.registry.JAXRException;
import javax.xml.registry.infomodel.Association;
import javax.xml.registry.infomodel.Concept;
import javax.xml.registry.infomodel.RegistryObject;
import javax.xml.registry.infomodel.Slot;

import net.sf.csutils.core.model.RORelation;
import net.sf.csutils.core.model.ROSlot;
import net.sf.csutils.core.registry.RegistryFacade;
import net.sf.csutils.core.registry.centrasite.CentraSiteRegistryInfo;
import net.sf.csutils.core.utils.RegistryObjects;

import com.centrasite.jaxr.type.CentraSiteSlotType;


public abstract class AbstractInnerModelAccessor implements InnerModelAccessor {
    protected String getSlotType(ROSlot pSlot) {
        switch (pSlot.getType()) {
            case bool:
                return CentraSiteSlotType.XS_BOOLEAN;
            case date:
                return CentraSiteSlotType.XS_DATE;
            case dateTime:
                return CentraSiteSlotType.XS_DATETIME;
            case duration:
                return CentraSiteSlotType.XS_DURATION;
            case emailAddress:
                return CentraSiteSlotType.EMAILADDRESS;
            case integer:
                return CentraSiteSlotType.XS_INTEGER;
            case ipv4Address:
                return CentraSiteSlotType.IPV4;
            case ipv6Address:
                return CentraSiteSlotType.IPV6;
            case longInteger:
                return CentraSiteSlotType.XS_INTEGER;
            case number:
                return CentraSiteSlotType.XS_DOUBLE;
            case string:
                return CentraSiteSlotType.XS_STRING;
            case uri:
                return CentraSiteSlotType.XS_ANYURI;
            case time:
                return CentraSiteSlotType.XS_TIME;
            default:
                throw new IllegalStateException("Invalid slot type: " + pSlot.getType());
        }
    }

    @Override
    public void setSlotValue(RegistryFacade pFacade, ROSlot pSlot, RegistryObject pObject, String pValue) throws JAXRException {
        if (pValue == null) {
            if (pSlot.getMinOccurs() > 0) {
                throw new JAXRException("Minimum number of values for slot "
                        + pSlot.getName() + " is " + pSlot.getMinOccurs());
            }
            final String slotName = getSlotName(pSlot.getName(), pObject);
            pObject.removeSlot(slotName);
        } else {
            if (pSlot.getMinOccurs() > 1) {
                throw new JAXRException("Minimum number of values for slot "
                        + pSlot.getName() + " is " + pSlot.getMinOccurs());
            }
            final String slotName = getSlotName(pSlot.getName(), pObject);
            final Slot slot = pFacade.getBusinessLifeCycleManager().createSlot(slotName, pValue, getSlotType(pSlot));
            pObject.addSlot(slot);
        }
    }

    private String getSlotName(String pProperty, final RegistryObject pObject)
            throws JAXRException {
        final String objectType = pObject.getObjectType().getValue();
        int offset = objectType.indexOf('}');
        final String slotName;
        if (offset == -1) {
            slotName = pProperty;
        } else {
            slotName = objectType.substring(0, offset+1) + pProperty;
        }
        return slotName;
    }

    @Override
    public void setSlotValues(RegistryFacade pFacade, ROSlot pSlot,
            RegistryObject pObject, Collection<String> pValues)
            throws JAXRException {
        final int size = pValues == null ? 0 : pValues.size();
        final int maxOccurs = pSlot.getMaxOccurs();
        if (maxOccurs != -1  &&  size > maxOccurs) {
            throw new JAXRException("Maximum number of values for slot "
                    + pSlot.getName() + " is " + maxOccurs);
        }
        if (size < pSlot.getMinOccurs()) {
            throw new JAXRException("Minimum number of values for slot "
                    + pSlot.getName() + " is " + pSlot.getMinOccurs());
        }
        final String slotName = getSlotName(pSlot.getName(), pObject);
        if (size == 0) {
            pObject.removeSlot(slotName);
        } else {
            final Slot slot = pFacade.getBusinessLifeCycleManager().createSlot(slotName, pValues, getSlotType(pSlot));
            pObject.addSlot(slot);
        }
    }

    @Override
    public String getSlotValue(RegistryFacade pFacade, ROSlot pSlot, RegistryObject pObject) throws JAXRException {
        final Slot slot = pObject.getSlot(getSlotName(pSlot.getName(), pObject));
        if (slot == null) {
            return null;
        }
        final Collection<?> values = slot.getValues();
        if (values.isEmpty()) {
            return null;
        }
        final Object o = values.iterator().next();
        return o == null ? null : o.toString();
    }

    @Override
    public List<String> getSlotValues(RegistryFacade pFacade, ROSlot pSlot,
            RegistryObject pObject) throws JAXRException {
        final Slot slot = pObject.getSlot(getSlotName(pSlot.getName(), pObject));
        if (slot == null) {
            return Collections.emptyList();
        }
        final Collection<?> values = slot.getValues();
        if (values.isEmpty()) {
            return Collections.emptyList();
        }
        final List<String> result = new ArrayList<String>();
        for (Object o : values) {
            result.add(o == null ? null : o.toString());
        }
        return result;
    }

    @Override
    public void setRelationValue(RegistryFacade pFacade, RORelation pRelation,
            RegistryObject pObject, RegistryObject pTarget)
            throws JAXRException {
        setRelationValues(pFacade, pRelation, pObject, Collections.singleton(pTarget));
    }

    @Override
    public void setRelationValues(RegistryFacade pFacade, RORelation pRelation,
            RegistryObject pObject, Collection<RegistryObject> pTargets)
            throws JAXRException {
        final Collection<Association> assocs = RegistryObjects.getAssociations(pObject);
        final String attributeKey = pRelation.getAttributeKey();
        if (attributeKey == null) {
            throw new JAXRException("The relations attribute key is not set in the model for relation "
                    + pRelation.getName() + " of type " + pObject.getObjectType().getValue());
        }
        final Concept concept = pFacade.findAssociationType(pRelation.getAssociationType());
        if (concept == null) {
            throw new JAXRException("No association type named " + pRelation.getName()
                    + " was found.");
        }
        final List<Association> assocsToRemove = new ArrayList<Association>();
        for (Association assoc : assocs) {
            final Slot slot = assoc.getSlot(CentraSiteRegistryInfo.SLOT_NAME_ATTRIBUTES_KEY);
            if (slot != null) {
                final Iterator<?> values = slot.getValues().iterator();
                if (values.hasNext()) {
                    final Object value = values.next();
                    if (attributeKey.equals(value)) {
                        assocsToRemove.add(assoc);
                    }
                }
            }
        }
        if (!assocsToRemove.isEmpty()) {
            pObject.removeAssociations(assocsToRemove);
        }
        final BusinessLifeCycleManager blcm = pFacade.getBusinessLifeCycleManager();
        for (RegistryObject target : pTargets) {
            final Association assoc = blcm.createAssociation(target, concept);
            pObject.addAssociation(assoc);
            final Slot slot = blcm.createSlot(CentraSiteRegistryInfo.SLOT_NAME_ATTRIBUTES_KEY, attributeKey, CentraSiteSlotType.XS_STRING);
            assoc.addSlot(slot);
        }
    }

    @Override
    public RegistryObject getRelationValue(RegistryFacade pFacade,
            RORelation pRelation, RegistryObject pObject) throws JAXRException {
        final Collection<Association> assocs = RegistryObjects.getAssociations(pObject);
        final String attributeKey = pRelation.getAttributeKey();
        if (attributeKey == null) {
            throw new JAXRException("The relations attribute key is not set in the model for relation "
                    + pRelation.getName() + " of type " + pObject.getObjectType().getValue());
        }
        for (Association assoc : assocs) {
            final Slot slot = assoc.getSlot(attributeKey);
            if (slot != null) {
                final Iterator<?> values = slot.getValues().iterator();
                if (values.hasNext()) {
                    final Object value = values.next();
                    if (attributeKey.equals(value)) {
                        return assoc.getTargetObject();
                    }
                }
            }
        }
        return null;
    }

    @Override
    public List<RegistryObject> getRelationValues(RegistryFacade pFacade,
            RORelation pRelation, RegistryObject pObject) throws JAXRException {
        final Collection<Association> assocs = RegistryObjects.getAssociations(pObject);
        final String attributeKey = pRelation.getAttributeKey();
        if (attributeKey == null) {
            throw new JAXRException("The relations attribute key is not set in the model for relation "
                    + pRelation.getName() + " of type " + pObject.getObjectType().getValue());
        }
        final List<RegistryObject> result = new ArrayList<RegistryObject>();
        for (Association assoc : assocs) {
            final Slot slot = assoc.getSlot(CentraSiteRegistryInfo.SLOT_NAME_ATTRIBUTES_KEY);
            if (slot != null) {
                final Iterator<?> values = slot.getValues().iterator();
                if (values.hasNext()) {
                    final Object value = values.next();
                    if (attributeKey.equals(value)) {
                        result.add(assoc.getTargetObject());
                    }
                }
            }
        }
        return result;
    }
}
