package net.sf.csutils.core.registry.impl;


import java.util.Collection;

import javax.xml.registry.BulkResponse;
import javax.xml.registry.JAXRException;
import javax.xml.registry.infomodel.Association;
import javax.xml.registry.infomodel.ClassificationScheme;
import javax.xml.registry.infomodel.Concept;
import javax.xml.registry.infomodel.RegistryObject;

import net.sf.csutils.core.model.QName;
import net.sf.csutils.core.registry.RegistryFacade;
import net.sf.csutils.core.registry.RegistryInfo;
import net.sf.csutils.core.utils.Generics;


/**
 * Abstract base class for deriving implementations of {@link RegistryFacade}.
 */
public abstract class AbstractRegistryFacade implements RegistryFacade {
	private final RegistryInfo registryInfo;

	/**
	 * Creates a new instance with the given {@link RegistryInfo}.
	 * @param pRegistryInfo The registry information object.
	 */
	public AbstractRegistryFacade(RegistryInfo pRegistryInfo) {
		registryInfo = pRegistryInfo;
	}

	/**
	 * Returns the {@link RegistryInfo}.
	 * @return The registry information object.
	 */
	@Override
	public RegistryInfo getRegistryInfo() {
		return registryInfo;
	}
	

	/**
	 * Returns the {@link Concept object type} with the given name.
	 */
    @Override
	public Concept findObjectType(QName pName) throws JAXRException {
		return findObjectType(pName.toString());
	}

	/**
	 * Returns, whether the given concept is an association type concept.
	 */
    @Override
	public boolean isAssociationTypeConcept(Concept pConcept) throws JAXRException {
		final Object parent = pConcept.getParent();
		if (parent instanceof ClassificationScheme) {
			return isAssociationTypeClassificationScheme((ClassificationScheme) parent);
		}
		return false;
	}

	/**
	 * Returns, whether the given concept is an object type concept.
	 */
    @Override
	public boolean isObjectTypeConcept(Concept pConcept) throws JAXRException {
		final Object parent = pConcept.getParent();
		if (parent instanceof ClassificationScheme) {
			return isObjectTypeClassificationScheme((ClassificationScheme) parent);
		}
		return false;
	}

	/**
	 * Returns, whether the given classification scheme is the association type classification scheme.
	 */
    @Override
	public boolean isAssociationTypeClassificationScheme(ClassificationScheme pClassificationScheme) throws JAXRException {
		return getRegistryInfo().getAssociationTypesKey().equals(pClassificationScheme.getKey().getId());
	}

	/**
	 * Returns, whether the given classification scheme is the object type classification scheme.
	 */
    @Override
	public boolean isObjectTypeClassificationScheme(ClassificationScheme pClassificationScheme) throws JAXRException {
		return getRegistryInfo().getObjectTypesKey().equals(pClassificationScheme.getKey().getId());
	}

	/**
	 * Returns the classification scheme of association types.
	 */
    @Override
	public ClassificationScheme getAssociationTypeClassificationScheme() throws JAXRException {
		return (ClassificationScheme) getBusinessQueryManager().getRegistryObject(getRegistryInfo().getAssociationTypesKey());
	}

	/**
	 * Returns the classification scheme of object types.
	 */
    @Override
	public ClassificationScheme getObjectTypeClassificationScheme() throws JAXRException {
		return (ClassificationScheme) getBusinessQueryManager().getRegistryObject(getRegistryInfo().getObjectTypesKey());
	}

	@Override
	public Collection<Association> getReferencingAssociations(RegistryObject pRo) throws JAXRException {
		final BulkResponse br = getBusinessQueryManager().findAssociations(null, null, pRo.getKey().getId(), null);
		return Generics.cast(br.getCollection());
	}
}
