package net.sf.csutils.core.registry.impl;

import static net.sf.csutils.core.utils.Generics.cast;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import javax.xml.registry.JAXRException;
import javax.xml.registry.infomodel.ClassificationScheme;
import javax.xml.registry.infomodel.Concept;

import net.sf.csutils.core.model.QName;
import net.sf.csutils.core.model.ROType;
import net.sf.csutils.core.registry.RegistryFacade;


/**
 * The meta model accessor is used to read, or modify the
 * registry meta model.
 */
public abstract class MetaModelAccessor {
    /**
     * Opaque implementation of a meta model.
     */
    public static class MetaModel {
        private final RegistryFacade facade;
        private final Map<QName,ROType> types = new HashMap<QName,ROType>(); 
        private Map<QName,ROType> typesCopy;
        private boolean haveAllTypes;

        public MetaModel(RegistryFacade pFacade) {
            facade = pFacade;
        }
        public Map<QName,ROType> getTypes() {
            return types;
        }
        public Map<QName,ROType> getTypesCopy() {
            return typesCopy;
        }
        public boolean hasAllTypes() {
            return haveAllTypes;
        }
        public RegistryFacade getFacade() {
            return facade;
        }
        public void clear() {
            types.clear();
            typesCopy = null;
            haveAllTypes = false;
        }
    }

    /**
     * Creates a meta model instance.
     * @param pFacade The facade to use.
     * @return The created meta model instance.
     */
    public abstract MetaModel newMetaModel(RegistryFacade pFacade);

    /**
     * Returns the {@link ROType registry object type} with the
     * given name.
     * @param pModel The meta model to modify.
     * @param pQName Qualified name of the registry object type to return.
     * @return The registry object type with the given name, if any, or null.
     * @throws JAXRException The operation failed.
     */
    public abstract ROType getROType(MetaModel pModel, QName pQName) throws JAXRException;

    /**
     * Removes the given {@link ROType registry object type} from
     * the registry.
     * @param pModel The inner meta model to modify.
     * @param pType The registry object type to remove.
     * @throws JAXRException The operation failed.
     */
    protected abstract void remove(MetaModel pModel, ROType pType) throws JAXRException;

    /**
     * Creates the given {@link ROType registry object type} in
     * the registry.
     * @param pModel The inner meta model to modify.
     * @param pType Description of the registry object type to create.
     * @return The created registry object type.
     * @throws JAXRException The operation failed.
     */
    protected abstract ROType add(MetaModel pModel, ROType pType) throws JAXRException;

    /**
     * Returns the map of all types.
     * @param pModel The inner meta model to use.
     * @return The map of all registry object types contained in the given meta model.
     * @throws JAXRException The operation failed.
     */
    public Map<QName, ROType> getROTypes(MetaModel pModel) throws JAXRException {
        synchronized (pModel) {
            if (pModel.haveAllTypes) {
                if (pModel.typesCopy == null) {
                    pModel.typesCopy = Collections.unmodifiableMap(pModel.types);
                }
                return pModel.typesCopy;
            }
        }
        final ClassificationScheme cs = pModel.facade.getObjectTypeClassificationScheme();
        final Collection<Concept> objectTypeConcepts = cast(cs.getChildrenConcepts());
        for (Concept concept : objectTypeConcepts) {
            // Fill the cache, by querying all the object types.
            final QName qName = QName.valueOf(concept.getValue());
            getROType(pModel, qName);
        }
        synchronized (pModel.types) {
            pModel.haveAllTypes = true;
            return getROTypes(pModel);
        }
    }

    /**
     * Clears the cached object types.
     * @param pModel The inner meta model to modify.
     * @throws JAXRException Clearing the cache didn't work.
     */
    public void clear(MetaModel pModel) throws JAXRException {
        synchronized(pModel) {
            pModel.clear();
        }
    }

    /**
     * Removes the given object type from the registry.
     * @param pModel The inner meta model to modify.
     * @param pQName The qualified name of the registry object type to remove.
     * @throws JAXRException The operation failed.
     */
    public void remove(MetaModel pModel, QName pQName) throws JAXRException {
        synchronized (pModel) {
            final ROType type = getROType(pModel, pQName);
            if (type == null) {
                throw new JAXRException("No such type: " + pQName);
            }
            remove(pModel, type);
            pModel.types.remove(pQName);
            pModel.typesCopy = null;
        }
    }

    /**
     * Adds a new object type to the registry.
     * @param pModel The inner meta model to modify.
     * @param pType Description of the registry object type to create.
     * @return The created registry object type.
     * @throws JAXRException The operation failed.
     */
    public ROType create(MetaModel pModel, ROType pType) throws JAXRException {
        synchronized (pModel) {
            if (pModel.types.get(pType.getQName()) != null) {
                throw new JAXRException("A type named " + pType.getQName()
                        + " already exists.");
            }
            final ROType type = add(pModel, pType);
            pModel.types.put(pType.getQName(), type);
            pModel.typesCopy = null;
            return type;
        }
    }
}
