package net.sf.csutils.core.registry.impl;

import java.util.Arrays;
import java.util.Collection;
import java.util.List;

import javax.xml.registry.JAXRException;
import javax.xml.registry.infomodel.RegistryObject;

import net.sf.csutils.core.model.QName;
import net.sf.csutils.core.model.ROAttribute;
import net.sf.csutils.core.model.ROMetaModel;
import net.sf.csutils.core.model.RORelation;
import net.sf.csutils.core.model.ROSlot;
import net.sf.csutils.core.model.ROType;
import net.sf.csutils.core.query.CsqlParser;
import net.sf.csutils.core.query.CsqlStatement;
import net.sf.csutils.core.query.StatementUtils;
import net.sf.csutils.core.registry.ModelDrivenRegistryFacade;
import net.sf.csutils.core.registry.ROMetaModelAccessor;
import net.sf.csutils.core.registry.ROModelAccessor;
import net.sf.csutils.core.registry.RegistryFacade;


public class ROModelAccessorImpl implements ROModelAccessor {
    private final ModelDrivenRegistryFacade facade;
    private final ROMetaModelAccessor metaModelAcc;
    private final InnerModelAccessor acc;

    public ROModelAccessorImpl(ModelDrivenRegistryFacade pFacade, ROMetaModelAccessor pMetaModelAccessor, InnerModelAccessor pAccessor) {
        facade = pFacade;
        metaModelAcc = pMetaModelAccessor;
        acc = pAccessor;
    }

    @Override
    public ROMetaModel getMetaModel() {
        return metaModelAcc.getMetaModel();
    }

    @Override
    public ModelDrivenRegistryFacade getRegistryFacade() {
        return facade;
    }

    @Override
    public ROType findType(QName pQName) throws JAXRException {
        return getMetaModel().getROType(pQName);
    }

    @Override
    public ROType requireType(QName pQName) throws JAXRException {
        final ROType type = findType(pQName);
        if (type == null) {
            throw new JAXRException("No type named " + pQName + " was found.");
        }
        return type;
    }

    @Override
    public ROType findType(String pTypeName) throws JAXRException {
        return findType(QName.valueOf(pTypeName));
    }

    @Override
    public ROType requireType(String pTypeName) throws JAXRException {
        return requireType(QName.valueOf(pTypeName));
    }

    @Override
    public ROType findType(RegistryObject pInstance) throws JAXRException {
        return findType(pInstance.getObjectType().getValue());
    }

    @Override
    public ROType requireType(RegistryObject pInstance) throws JAXRException {
        return requireType(pInstance.getObjectType().getValue());
    }

    @Override
    public RORelation findRelation(ROType pType, String pRelationName)
            throws JAXRException {
        final ROAttribute attr = pType.getAttribute(pRelationName);
        if (attr == null) {
            return null;
        }
        if (attr instanceof RORelation) {
            return (RORelation) attr;
        }
        throw new JAXRException("The attribute " + pRelationName
                + " of type " + pType.getQName() + " is no relationship attribute.");
    }

    @Override
    public ROSlot findSlot(ROType pType, String pSlotName) throws JAXRException {
        final ROAttribute attr = pType.getAttribute(pSlotName);
        if (attr == null) {
            return null;
        }
        if (attr instanceof ROSlot) {
            return (ROSlot) attr;
        }
        throw new JAXRException("The attribute " + pSlotName
                + " of type " + pType.getQName() + " is no slot attribute.");
    }

    @Override
    public ROSlot requireSlot(ROType pType, String pSlotName) throws JAXRException {
        final ROSlot slot = findSlot(pType, pSlotName);
        if (slot == null) {
            throw new JAXRException("Unknown attribute " + pSlotName
                    + " for type " + pType.getQName());
        }
        return slot;
    }

    @Override
    public RORelation requireRelation(ROType pType, String pRelationName) throws JAXRException {
        final RORelation relation = findRelation(pType, pRelationName);
        if (relation == null) {
            throw new JAXRException("Unknown attribute " + pRelationName
                    + " for type " + pType.getQName());
        }
        return relation;
    }


    protected ROSlot findSlot(ROMetaModel pModel, RegistryObject pObject, String pSlotName) throws JAXRException {
        final QName qName = QName.valueOf(pObject.getObjectType().getValue());
        final ROType type = pModel.getROType(qName);
        if (type == null) {
            throw new JAXRException("The object type " + pObject.getObjectType().getValue()
                    + " is not known in the model.");
        }
        return findSlot(type, pSlotName);
    }

    @Override
    public void setSlotValue(ROSlot pSlot, RegistryObject pObject, String pValue)
            throws JAXRException {
        acc.setSlotValue(facade, pSlot, pObject, pValue);
    }

    @Override
    public void setSlotValue(ROType pType, String pSlotName,
            RegistryObject pObject, String pValue) throws JAXRException {
        acc.setSlotValue(facade, requireSlot(pType, pSlotName), pObject, pValue);
    }

    @Override
    public void setSlotValue(String pSlotName, RegistryObject pObject,
            String pValue) throws JAXRException {
        acc.setSlotValue(facade, requireSlot(requireType(pObject), pSlotName), pObject, pValue);
    }

    @Override
    public void setSlotValues(ROSlot pSlot, RegistryObject pObject,
            Collection<String> pValues) throws JAXRException {
        acc.setSlotValues(facade, pSlot, pObject, pValues);
    }

    @Override
    public void setSlotValues(ROType pType, String pSlotName,
            RegistryObject pObject, Collection<String> pValues)
            throws JAXRException {
        acc.setSlotValues(facade, requireSlot(pType, pSlotName), pObject, pValues);
    }

    @Override
    public void setSlotValues(String pSlotName, RegistryObject pObject,
            Collection<String> pValues) throws JAXRException {
        acc.setSlotValues(facade, requireSlot(requireType(pObject), pSlotName), pObject, pValues);
    }

    @Override
    public void setSlotValues(ROSlot pSlot, RegistryObject pObject,
            String... pValues) throws JAXRException {
        acc.setSlotValues(facade, pSlot, pObject, Arrays.asList(pValues));
    }

    @Override
    public void setSlotValues(ROType pType, String pSlotName,
            RegistryObject pObject, String... pValues) throws JAXRException {
        acc.setSlotValues(facade, requireSlot(pType, pSlotName), pObject, Arrays.asList(pValues));
    }

    @Override
    public void setSlotValues(String pSlotName, RegistryObject pObject,
            String... pValues) throws JAXRException {
        acc.setSlotValues(facade, requireSlot(requireType(pObject), pSlotName), pObject, Arrays.asList(pValues));
    }

    @Override
    public String getSlotValue(ROSlot pSlot, RegistryObject pObject)
            throws JAXRException {
        return acc.getSlotValue(facade, pSlot, pObject);
    }

    @Override
    public String getSlotValue(ROType pType, String pSlotName,
            RegistryObject pObject) throws JAXRException {
        return acc.getSlotValue(facade, requireSlot(pType, pSlotName), pObject);
    }

    @Override
    public String getSlotValue(String pSlotName, RegistryObject pObject)
            throws JAXRException {
        return acc.getSlotValue(facade, requireSlot(requireType(pObject), pSlotName), pObject);
    }

    @Override
    public Collection<String> getSlotValues(ROSlot pSlot, RegistryObject pObject)
            throws JAXRException {
        return acc.getSlotValues(facade, pSlot, pObject);
    }

    @Override
    public Collection<String> getSlotValues(ROType pType, String pSlotName,
            RegistryObject pObject) throws JAXRException {
        return acc.getSlotValues(facade, requireSlot(pType, pSlotName), pObject);
    }

    @Override
    public Collection<String> getSlotValues(String pSlotName,
            RegistryObject pObject) throws JAXRException {
        return acc.getSlotValues(facade, requireSlot(requireType(pObject), pSlotName), pObject);
    }

    @Override
    public void setRelationValue(RORelation pRelation, RegistryObject pObject,
            RegistryObject pTarget) throws JAXRException {
        acc.setRelationValue(facade, pRelation, pObject, pTarget);
    }

    @Override
    public void setRelationValue(ROType pType, String pRelationName,
            RegistryObject pObject, RegistryObject pTarget) throws JAXRException {
        acc.setRelationValue(facade, requireRelation(pType, pRelationName), pObject, pTarget);
    }

    @Override
    public void setRelationValue(String pRelationName, RegistryObject pObject,
            RegistryObject pTarget) throws JAXRException {
        acc.setRelationValue(facade, requireRelation(requireType(pObject), pRelationName), pObject, pTarget);
    }

    @Override
    public void setRelationValues(RORelation pRelation, RegistryObject pObject,
            RegistryObject... pTargets) throws JAXRException {
        acc.setRelationValues(facade, pRelation, pObject, Arrays.asList(pTargets));
    }

    @Override
    public void setRelationValues(ROType pType, String pRelationName,
            RegistryObject pObject, RegistryObject... pTargets)
            throws JAXRException {
        acc.setRelationValues(facade, requireRelation(pType, pRelationName), pObject, Arrays.asList(pTargets));
    }

    @Override
    public void setRelationValues(String pRelationName, RegistryObject pObject,
            RegistryObject... pTargets) throws JAXRException {
        acc.setRelationValues(facade, requireRelation(requireType(pObject), pRelationName), pObject, Arrays.asList(pTargets));
    }

    @Override
    public void setRelationValues(RORelation pRelation, RegistryObject pObject,
            Collection<RegistryObject> pTargets) throws JAXRException {
        acc.setRelationValues(facade, pRelation, pObject, pTargets);
    }

    @Override
    public void setRelationValues(ROType pType, String pRelationName,
            RegistryObject pObject, Collection<RegistryObject> pTargets)
            throws JAXRException {
        acc.setRelationValues(facade, requireRelation(pType, pRelationName), pObject, pTargets);
    }

    @Override
    public void setRelationValues(String pRelationName, RegistryObject pObject,
            Collection<RegistryObject> pTargets) throws JAXRException {
        acc.setRelationValues(facade, requireRelation(requireType(pObject), pRelationName), pObject, pTargets);
    }

    @Override
    public RegistryObject getRelationValue(RORelation pRelation,
            RegistryObject pObject) throws JAXRException {
        return acc.getRelationValue(facade, pRelation, pObject);
    }

    @Override
    public RegistryObject getRelationValue(ROType pType, String pRelationName,
            RegistryObject pObject) throws JAXRException {
        return acc.getRelationValue(facade, requireRelation(pType, pRelationName), pObject);
    }

    @Override
    public RegistryObject getRelationValue(String pRelationName,
            RegistryObject pObject) throws JAXRException {
        return acc.getRelationValue(facade, requireRelation(requireType(pObject), pRelationName), pObject);
    }

    @Override
    public Collection<RegistryObject> getRelationValues(RORelation pRelation,
            RegistryObject pObject) throws JAXRException {
        return acc.getRelationValues(facade, pRelation, pObject);
    }

    @Override
    public Collection<RegistryObject> getRelationValues(ROType pType,
            String pRelationName, RegistryObject pObject) throws JAXRException {
        return acc.getRelationValues(facade, requireRelation(pType, pRelationName), pObject);
    }

    @Override
    public Collection<RegistryObject> getRelationValues(String pRelationName,
            RegistryObject pObject) throws JAXRException {
        return acc.getRelationValues(facade, requireRelation(requireType(pObject), pRelationName), pObject);
    }


    @Override
    @SuppressWarnings("unchecked")
    public <RO extends RegistryObject> RO createInstance(ROType pType) throws JAXRException {
        return (RO) createInstance(pType.getQName());
    }

    @Override
    @SuppressWarnings("unchecked")
    public <RO extends RegistryObject> RO createInstance(QName pType) throws JAXRException {
        return (RO) createInstance(pType.toString());
    }

    @Override
    @SuppressWarnings("unchecked")
    public <RO extends RegistryObject> RO createInstance(String pType) throws JAXRException {
        return (RO) acc.createInstance(getRegistryFacade(), pType);
    }

    @Override
    public CsqlStatement prepareStatement(String pQuery) throws JAXRException {
        final RegistryFacade fcd = getRegistryFacade();
        return new CsqlParser(getMetaModel(), fcd.getRegistryInfo()).parse(fcd, pQuery);
    }

    @Override
    public List<RegistryObject> executeQuery(String pQuery,
            Object... pParameters) throws JAXRException {
        final CsqlStatement stmt = prepareStatement(pQuery);
        StatementUtils.setParameters(stmt, pParameters);
        return stmt.executeQuery();
    }

    @Override
    public List<RegistryObject[]> executeArrayQuery(String pQuery,
            Object... pParameters) throws JAXRException {
        final CsqlStatement stmt = prepareStatement(pQuery);
        StatementUtils.setParameters(stmt, pParameters);
        return stmt.executeArrayQuery();
    }
}
