package net.sf.csutils.core.utils;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.Reader;
import java.lang.reflect.UndeclaredThrowableException;
import java.net.URL;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;
import javax.xml.stream.XMLEventReader;
import javax.xml.stream.XMLStreamReader;
import javax.xml.transform.Source;
import javax.xml.transform.sax.SAXSource;
import javax.xml.transform.stream.StreamSource;

import org.w3c.dom.Node;
import org.xml.sax.InputSource;


/**
 * Helper class, which provides useful tools for working with JAXB.
 */
public class JaxbReader<CLSS> {
	private final JAXBContext ctx;
	private final Class<CLSS> theClass;

	public JaxbReader(Class<CLSS> pClass) {
		theClass = pClass;
		try {
			ctx = JAXBContext.newInstance(pClass);
		} catch (JAXBException e) {
			throw new UndeclaredThrowableException(e);
		}
	}
	
	/**
	 * Reads and deserializes a JAXB bean by reading the given {@link InputStream}.
	 * @param pStream The stream, from which to read the JAXB bean.
	 * @return The deserialized JAXB bean.
	 * @throws JAXBException The operation failed.
	 */
	public CLSS read(InputStream pStream) throws JAXBException {
		return read(new StreamSource(pStream));
	}

	/**
	 * Reads and deserializes a JAXB bean from the XML stream provided
	 * by the given {@link File}.
     * @param pFile The file, from which to read the JAXB bean.
	 * @return The deserialized JAXB bean.
     * @throws JAXBException The operation failed.
	 */
	public CLSS read(File pFile) throws JAXBException {
		return read(new StreamSource(pFile));
	}

	/**
	 * Reads and deserializes a JAXB bean from the XML stream provided
	 * by the given {@link InputSource}.
     * @param pSource The SAX source, from which to read the JAXB bean.
	 * @return The deserialized JAXB bean.
     * @throws JAXBException The operation failed.
	 */
	public CLSS read(InputSource pSource) throws JAXBException {
		return read(new SAXSource(pSource));
	}

	/**
	 * Reads and deserializes a JAXB bean from the XML stream provided
	 * by the given {@link Reader}.
     * @param pReader The reader, from which to read the JAXB bean.
	 * @return The deserialized JAXB bean.
     * @throws JAXBException The operation failed.
	 */
	public CLSS read(Reader pReader) throws JAXBException {
		return read(new StreamSource(pReader));
	}

	/**
	 * Reads and deserializes a JAXB bean from the XML stream provided
	 * by the given {@link Node}.
     * @param pNode The node, from which to read the XML stream.
	 * @return The deserialized JAXB bean.
	 * @throws JAXBException The operation failed.
	 */
	public CLSS read(Node pNode) throws JAXBException {
		final JAXBElement<CLSS> bean = ctx.createUnmarshaller().unmarshal(pNode, theClass);
		return bean.getValue();
	}

	/**
	 * Reads and deserializes a JAXB bean from the XML stream provided
	 * by the given {@link Source}.
     * @param pSource The transformer source providing the XML stream.
	 * @return The deserialized JAXB bean.
     * @throws JAXBException The operation failed.
	 */
	public CLSS read(Source pSource) throws JAXBException {
		final JAXBElement<CLSS> bean = ctx.createUnmarshaller().unmarshal(pSource, theClass);
		return bean.getValue();
	}

	/**
	 * Reads and deserializes a JAXB bean from the XML stream provided
	 * by the given {@link URL}.
     * @param pURL The URL providing the XML streams.
	 * @return The deserialized JAXB bean.
     * @throws JAXBException The operation failed.
	 */
	public CLSS read(URL pURL) throws IOException, JAXBException {
		InputStream stream = null;
		try {
			stream = pURL.openStream();
			final InputSource inputSource = new InputSource(stream);
			inputSource.setSystemId(pURL.toExternalForm());
			final CLSS result = read(inputSource);
			stream.close();
			stream = null;
			return result;
		} finally {
			if (stream != null) { try { stream.close(); } catch (Throwable t) { /* Igmore me */ } }
		}
	}

	/**
	 * Reads and deserializes a JAXB bean from the XML stream provided
	 * by the given {@link XMLEventReader}.
     * @param pReader The StAX event reader poviding the XML stream.
	 * @return The deserialized JAXB bean.
     * @throws JAXBException The operation failed.
	 */
	public CLSS read(XMLEventReader pReader) throws JAXBException {
		final JAXBElement<CLSS> bean = ctx.createUnmarshaller().unmarshal(pReader, theClass);
		return bean.getValue();
	}

	/**
	 * Reads and deserializes a JAXB bean from the XML stream provided
	 * by the given {@link XMLStreamReader}.
     * @param pReader The StAX stream reader providing the XML stream.
	 * @return The deserialized JAXB bean.
     * @throws JAXBException The operation failed.
	 */
	public CLSS read(XMLStreamReader pReader) throws JAXBException {
		final JAXBElement<CLSS> bean = ctx.createUnmarshaller().unmarshal(pReader, theClass);
		return bean.getValue();
	}

	public CLSS read(Object pObject) throws JAXBException, IOException {
		if (pObject == null) {
			throw new IllegalArgumentException("The object to read must not be null.");
		}
		if (pObject instanceof File) {
			return read((File) pObject);
		}
		if (pObject instanceof InputStream) {
			return read((InputStream) pObject);
		}
		if (pObject instanceof InputSource) {
			return read((InputSource) pObject);
		}
		if (pObject instanceof Node) {
			return read((Node) pObject);
		}
		if (pObject instanceof Reader) {
			return read((Reader) pObject);
		}
		if (pObject instanceof Source) {
			return read((Source) pObject);
		}
		if (pObject instanceof URL) {
			return read((URL) pObject);
		}
		if (pObject instanceof XMLEventReader) {
			return read((XMLEventReader) pObject);
		}
		if (pObject instanceof XMLStreamReader) {
			return read((XMLStreamReader) pObject);
		}
		throw new IllegalArgumentException("Invalid object type: " + pObject.getClass().getName());
	}
}
