package net.sf.csutils.core.utils;

import java.io.File;
import java.io.OutputStream;
import java.io.Writer;
import java.lang.reflect.UndeclaredThrowableException;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;
import javax.xml.namespace.QName;
import javax.xml.stream.XMLEventWriter;
import javax.xml.stream.XMLStreamWriter;
import javax.xml.transform.Result;

import org.w3c.dom.Node;
import org.xml.sax.ContentHandler;


/**
 * The JAXB bean writer is used to serialize
 * JAXB beans models into XML.
 */
public class JaxbWriter<CLSS> {
	private final Class<CLSS> theClass;
	private final JAXBContext ctx;
	private final QName qName;

	public JaxbWriter(Class<CLSS> pClass, QName pQName) {
		try {
			theClass = pClass;
			ctx = JAXBContext.newInstance(pClass);
			qName = pQName;
		} catch (JAXBException e) {
			throw new UndeclaredThrowableException(e);
		}
	}
	

	/**
	 * Converts the given JAXB bean to XML by writing to the given
	 * {@link OutputStream}.
	 * @param pObject The JAXB bean to convert.
	 * @param pStream The stream to write to.
	 * @throws JAXBException The operation failed.
	 */
	public void write(CLSS pObject, OutputStream pStream) throws JAXBException {
		ctx.createMarshaller().marshal(asElement(pObject), pStream);
	}

	/**
	 * Converts the given JAXB bean to XML by writing to the given
	 * {@link File}.
     * @param pObject The JAXB bean to convert.
     * @param pFile The file to write to.
     * @throws JAXBException The operation failed.
	 */
	public void write(CLSS pObject, File pFile) throws JAXBException {
		ctx.createMarshaller().marshal(asElement(pObject), pFile);
	}

	/**
	 * Converts the given JAXB bean to XML by writing to the given
	 * {@link Writer}.
     * @param pObject The JAXB bean to convert.
     * @param pWriter The writer to write to.
     * @throws JAXBException The operation failed.
	 */
	public void write(CLSS pObject, Writer pWriter) throws JAXBException {
		ctx.createMarshaller().marshal(asElement(pObject), pWriter);
	}

	private JAXBElement<CLSS> asElement(final CLSS pBean) {
		return new JAXBElement<CLSS>(qName, theClass, pBean);
	}

	/**
	 * Converts the given JAXB bean to XML by writing to the given
	 * {@link Result}.
     * @param pObject The JAXB bean to convert.
     * @param pResult The transformer result to write to.
     * @throws JAXBException The operation failed.
	 */
	public void write(CLSS pObject, Result pResult) throws JAXBException {
		ctx.createMarshaller().marshal(asElement(pObject), pResult);
	}

	/**
	 * Converts the given JAXB bean to XML by writing to the given
	 * {@link Node}.
     * @param pObject The JAXB bean to convert.
     * @param pNode The node to write to.
     * @throws JAXBException The operation failed.
	 */
	public void write(CLSS pObject, Node pNode) throws JAXBException {
		ctx.createMarshaller().marshal(asElement(pObject), pNode);
	}

	/**
	 * Converts the given JAXB bean to XML by writing to the given
	 * {@link XMLEventWriter}.
     * @param pObject The JAXB bean to convert.
     * @param pWriter The StAX event writer to write to.
     * @throws JAXBException The operation failed.
	 */
	public void write(CLSS pObject, XMLEventWriter pWriter) throws JAXBException {
		ctx.createMarshaller().marshal(asElement(pObject), pWriter);
	}

	/**
	 * Converts the given JAXB bean to XML by writing to the given
	 * {@link XMLStreamWriter}.
     * @param pObject The JAXB bean to convert.
     * @param pWriter The StAX stream writer to write to.
     * @throws JAXBException The operation failed.
	 */
	public void write(CLSS pObject, XMLStreamWriter pWriter) throws JAXBException {
		ctx.createMarshaller().marshal(asElement(pObject), pWriter);
	}

	/**
	 * Converts the given JAXB bean to XML by writing to the given
	 * {@link ContentHandler}.
     * @param pObject The JAXB bean to convert.
     * @param pHandler The SAX content handler to write to.
     * @throws JAXBException The operation failed.
	 */
	public void write(CLSS pObject, ContentHandler pHandler) throws JAXBException {
		ctx.createMarshaller().marshal(asElement(pObject), pHandler);
	}

	/** Converts the given JAXB bean to XML by writing to the given output object.
	 * @param pObject The JAXB bean to convert.
	 * @param pOutput The output object
	 * @throws JAXBException The operation failed.
	 */
	public void write(CLSS pObject, Object pOutput) throws JAXBException {
		if (pOutput == null) {
			throw new IllegalArgumentException("The output object must not be null.");
		}
		if (pOutput instanceof ContentHandler) {
			write(pObject, (ContentHandler) pOutput);
		} else if (pOutput instanceof File) {
			write(pObject, (File) pOutput);
		} else if (pOutput instanceof Node) {
			write(pObject, (Node) pOutput);
		} else if (pOutput instanceof OutputStream) {
			write(pObject, (OutputStream) pOutput);
		} else if (pOutput instanceof Result) {
			write(pObject, (Result) pOutput);
		} else if (pOutput instanceof Writer) {
			write(pObject, (Writer) pOutput);
		} else if (pOutput instanceof XMLEventWriter) {
			write(pObject, (XMLEventWriter) pOutput);
		} else if (pOutput instanceof XMLStreamWriter) {
			write(pObject, (XMLStreamWriter) pOutput);
		} else {
			throw new IllegalArgumentException("Invalid output object type: "
				+ pOutput.getClass().getName());
		}
	}
}
