package net.sf.csutils.core.utils;

import java.util.Collection;
import java.util.Iterator;
import java.util.Locale;
import java.util.StringTokenizer;

import javax.xml.registry.JAXRException;
import javax.xml.registry.infomodel.Association;
import javax.xml.registry.infomodel.Classification;
import javax.xml.registry.infomodel.InternationalString;
import javax.xml.registry.infomodel.LocalizedString;
import javax.xml.registry.infomodel.RegistryObject;


/**
 * Utility method for working with registry objects.
 */
public class RegistryObjects {
	/**
	 * Returns the registry objects name. Attempts the following methods to
	 * find a name:
	 * <ol>
	 *   <li>Calls the {@link InternationalString#getValue()} method and uses
	 *     the result, if it is non-null, and non-empty.</li>
	 *   <li>Calls the {@link InternationalString#getValue(Locale)} method with
	 *     the {@link Locale#US} locale and uses the result, if it is non-null,
	 *     and non-empty.</li>
     *   <li>Calls the {@link InternationalString#getValue(Locale)} method with
     *     the {@link Locale#ENGLISH} locale and uses the result, if it is non-null,
     *     and non-empty.</li>
     *   <li>Calls the {@link InternationalString#getLocalizedStrings()} method
     *     and uses the first non-null, and non-empty value that is found.</li>
	 * </ol>
	 * @param pRo The registry object to query for its name.
	 * @return The registry objects name, if any, or null.
	 * @throws JAXRException The operation failed.
	 */
	public static String getName(RegistryObject pRo) throws JAXRException {
		return getValue(pRo.getName());
	}

	/**
	 * Returns the registry objects name with the given Locale.
	 * @param pRo The registry object to query for its name.
	 * @param pLocale The locales name as a string; internally
	 *   {@link #asLocale(String)} is invoked to convert the string
	 *   into a locale.
	 * @return The registry objects name with the given locale, if any, or null.
	 * @throws JAXRException The operation failed.
	 */
	public static String getName(RegistryObject pRo, String pLocale) throws JAXRException {
		return getValue(pRo.getName(), pLocale);
	}

	/**
	 * Returns the registry objects name with the given Locale.
     * @param pRo The registry object to query for its name.
     * @param pLocale The locales name.
     * @return The registry objects name with the given locale, if any, or null.
     * @throws JAXRException The operation failed.
	 */
	public static String getName(RegistryObject pRo, Locale pLocale) throws JAXRException {
		return getValue(pRo.getName(), pLocale);
	}

	/**
     * Returns the registry objects description. Attempts the following methods to
     * find a description:
     * <ol>
     *   <li>Calls the {@link InternationalString#getValue()} method and uses
     *     the result, if it is non-null, and non-empty.</li>
     *   <li>Calls the {@link InternationalString#getValue(Locale)} method with
     *     the {@link Locale#US} locale and uses the result, if it is non-null,
     *     and non-empty.</li>
     *   <li>Calls the {@link InternationalString#getValue(Locale)} method with
     *     the {@link Locale#ENGLISH} locale and uses the result, if it is non-null,
     *     and non-empty.</li>
     *   <li>Calls the {@link InternationalString#getLocalizedStrings()} method
     *     and uses the first non-null, and non-empty value that is found.</li>
     * </ol>
     * @param pRo The registry object to query for its description.
     * @return The registry objects description, if any, or null.
     * @throws JAXRException The operation failed.
	 */
	public static String getDescription(RegistryObject pRo) throws JAXRException {
		return getValue(pRo.getDescription());
	}

    /**
     * Returns the registry objects description with the given Locale.
     * @param pRo The registry object to query for its description.
     * @param pLocale The locales name as a string; internally
     *   {@link #asLocale(String)} is invoked to convert the string
     *   into a locale.
     * @return The registry objects description with the given locale, if any, or null.
     * @throws JAXRException The operation failed.
     */
	public static String getDescription(RegistryObject pRo, String pLocale) throws JAXRException {
		return getValue(pRo.getDescription(), pLocale);
	}

    /**
     * Returns the registry objects description with the given Locale.
     * @param pRo The registry object to query for its description.
     * @param pLocale The locales name.
     * @return The registry objects description with the given locale, if any, or null.
     * @throws JAXRException The operation failed.
     */
	public static String getDescription(RegistryObject pRo, Locale pLocale) throws JAXRException {
		return getValue(pRo.getDescription(), pLocale);
	}

	private static String getValue(InternationalString pValue) throws JAXRException {
		String value = pValue.getValue();
		if (value == null) {
			value = pValue.getValue(Locale.US);
			if (value == null) {
				value = pValue.getValue(Locale.ENGLISH);
				if (value == null) {
					final Collection<LocalizedString> values = Generics.cast(pValue.getLocalizedStrings());
					final Iterator<LocalizedString> iter = values.iterator();
					value = iter.hasNext() ? iter.next().getValue() : null;
				}
			}
		}
		return value;
	}

	private static String getValue(InternationalString pValue, Locale pLocale) throws JAXRException {
		return pValue.getValue(pLocale);
	}

	private static String getValue(InternationalString pValue, String pLocale) throws JAXRException {
		return pValue.getValue(asLocale(pLocale));
	}

	/**
	 * Converts the given string into a locale.
	 * @param pLocale The locales name; use the hyphen ('-') or underscore ('_')
	 *   characters as separators for language, country, and variant.
	 * @return The locale with the given name.
	 */
	public static Locale asLocale(String pLocale) {
		if (pLocale == null) {
			throw new IllegalArgumentException("Invalid locale: " + pLocale);
		}
		final StringTokenizer st = new StringTokenizer(pLocale, "-_");
		if (st.hasMoreTokens()) {
			final String language = st.nextToken();
			if (st.hasMoreTokens()) {
				final String country = st.nextToken();
				if (st.hasMoreTokens()) {
					final String variant = st.nextToken();
					return new Locale(language, country, variant);
				} else {
					return new Locale(language, country);
				}
			} else {
				return new Locale(language);
			}
		} else {
			throw new IllegalArgumentException("Invalid locale: " + pLocale);
		}
	}

    /**
     * Returns the given objects associations. Shortcut for
     * <pre>
     *   (Collection<Association>) pObject.getAssociations()
     * </pre>
     * @param pObject The registry object to query for associations.
     * @return The associations of the given registry object.
     * @throws JAXRException The operation failed.
     */
    @SuppressWarnings("unchecked")
    public static Collection<Association> getAssociations(RegistryObject pObject) throws JAXRException {
    	return pObject.getAssociations();
    }

    /**
     * Returns the given objects classifications. Shortcut for
     * <pre>
     *   (Collection<Classification>) pObject.getClassifications()
     * </pre>
     * @param pObject The registry object to query for classifications.
     * @return The classifications of the given registry object.
     * @throws JAXRException The operation failed.
     */
    @SuppressWarnings("unchecked")
    public static Collection<Classification> getClassifications(RegistryObject pObject) throws JAXRException {
    	return pObject.getClassifications();
    }
}
